"""Socket connection."""
import logging
import socket

from ..connections import StickConnection
from ..exceptions import PortError

_LOGGER = logging.getLogger(__name__)


class SocketConnection(StickConnection):
    """Wrapper for Socket connection configuration."""

    def __init__(self, port, parser):
        super().__init__(port, parser)
        # get the address from a <host>:<port> format
        port_split = self.port.split(":")
        self._socket_host = port_split[0]
        self._socket_port = int(port_split[1])
        self._socket_address = (self._socket_host, self._socket_port)

        self._socket = None

    def _open_connection(self):
        """Open socket."""
        _LOGGER.debug(
            "Open socket to host '%s' at port %s",
            self._socket_host,
            str(self._socket_port),
        )
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.connect(self._socket_address)
        except Exception as err:
            _LOGGER.debug(
                "Failed to connect to host %s at port %s, %s",
                self._socket_host,
                str(self._socket_port),
                err,
            )
            raise PortError(err)
        else:
            self._reader_start("socket_reader_thread")
            self._writer_start("socket_writer_thread")
            self._is_connected = True
            _LOGGER.debug(
                "Successfully connected to host '%s' at port %s",
                self._socket_host,
                str(self._socket_port),
            )

    def _close_connection(self):
        """Close the socket."""
        try:
            self._socket.close()
        except Exception as err:
            _LOGGER.debug(
                "Failed to close socket to host %s at port %s, %s",
                self._socket_host,
                str(self._socket_port),
                err,
            )
            raise PortError(err)

    def _read_data(self):
        """Read data from socket."""
        if self._is_connected:
            try:
                socket_data = self._socket.recv(9999)
            except Exception as err:
                _LOGGER.debug(
                    "Error while reading data from host %s at port %s : %s",
                    self._socket_host,
                    str(self._socket_port),
                    err,
                )
                self._is_connected = False
                raise PortError(err)
            else:
                return socket_data
        return None

    def _write_data(self, data):
        """Write data to socket."""
        try:
            self._socket.send(data)
        except Exception as err:
            _LOGGER.debug("Error while writing data to socket port : %s", err)
            self._is_connected = False
            raise PortError(err)



"""
Message controller for USB-Stick

The controller will:
- handle the connection (connect/disconnect) to the USB-Stick
- take care for message acknowledgements based on sequence id's
- resend message requests when timeouts occurs
- holds a sending queue and submit messages based on the message priority (high, medium, low)
- passes received messages back to message processor (stick.py)
- execution of callbacks after processing the response message

"""

from datetime import datetime, timedelta
import logging
from queue import Empty, PriorityQueue
import threading
import time

from .connections.serial import PlugwiseUSBConnection
from .connections.socket import SocketConnection
from .constants import (
    MESSAGE_RETRY,
    MESSAGE_TIME_OUT,
    PRIORITY_MEDIUM,
    REQUEST_FAILED,
    REQUEST_SUCCESS,
    SLEEP_TIME,
    STATUS_RESPONSES,
    UTF8_DECODE,
)
from .messages.requests import NodeInfoRequest, NodePingRequest, NodeRequest
from .messages.responses import (
    NodeAckLargeResponse,
    NodeAckResponse,
    NodeAckSmallResponse,
)
from .parser import PlugwiseParser
from .util import inc_seq_id

_LOGGER = logging.getLogger(__name__)


class StickMessageController:
    """Handle connection and message sending and receiving"""

    def __init__(self, port: str, message_processor, node_state):
        """Initialize message controller """
        self.connection = None
        self.discovery_finished = False
        self.expected_responses = {}
        self.init_callback = None
        self.last_seq_id = None
        self.message_processor = message_processor
        self.node_state = node_state
        self.parser = PlugwiseParser(self.message_handler)
        self.port = port

        self._send_message_queue = None
        self._send_message_thread = None
        self._receive_timeout_thread = False
        self._run_receive_timeout_thread = False
        self._run_send_message_thread = False

    def connect_to_stick(self, callback=None) -> bool:
        """
        Connect to USB-Stick and startup all worker threads

        Return: True when connection is successful.
        """
        self.init_callback = callback
        # Open connection to USB Stick
        if ":" in self.port:
            _LOGGER.debug(
                "Open socket connection to %s hosting Plugwise USB stick", self.port
            )
            self.connection = SocketConnection(self.port, self.parser.feed)
        else:
            _LOGGER.debug("Open USB serial connection to Plugwise USB stick")
            self.connection = PlugwiseUSBConnection(self.port, self.parser.feed)
        if self.connection.connect():
            _LOGGER.debug("Starting message controller threads...")
            # send daemon
            self._send_message_queue = PriorityQueue()
            self._run_send_message_thread = True
            self._send_message_thread = threading.Thread(
                None, self._send_message_loop, "send_messages_thread", (), {}
            )
            self._send_message_thread.daemon = True
            self._send_message_thread.start()
            # receive timeout daemon
            self._run_receive_timeout_thread = True
            self._receive_timeout_thread = threading.Thread(
                None, self._receive_timeout_loop, "receive_timeout_thread", (), {}
            )
            self._receive_timeout_thread.daemon = True
            self._receive_timeout_thread.start()
            _LOGGER.debug("All message controller threads started")
        else:
            _LOGGER.warning("Failed to connect to USB stick")
        return self.connection.is_connected()

    def send(
        self,
        request: NodeRequest,
        callback=None,
        retry_counter=0,
        priority=PRIORITY_MEDIUM,
    ):
        """Queue request message to be sent into Plugwise Zigbee network."""
        _LOGGER.debug(
            "Queue %s to be send with retry counter %s and priority %s",
            request.__class__.__name__,
            str(retry_counter),
            str(priority),
        )
        self._send_message_queue.put(
            (
                priority,
                retry_counter,
                datetime.now(),
                [
                    request,
                    callback,
                    retry_counter,
                    None,
                ],
            )
        )

    def resend(self, seq_id):
        """Resend message."""
        _mac = "<unknown>"
        if not self.expected_responses.get(seq_id):
            _LOGGER.warning(
                "Cannot resend unknown request %s",
                str(seq_id),
            )
        else:
            if self.expected_responses[seq_id][0].mac:
                _mac = self.expected_responses[seq_id][0].mac.decode(UTF8_DECODE)
            _request = self.expected_responses[seq_id][0].__class__.__name__

            if self.expected_responses[seq_id][2] == -1:
                _LOGGER.debug("Drop single %s to %s ", _request, _mac)
            elif self.expected_responses[seq_id][2] <= MESSAGE_RETRY:
                if (
                    isinstance(self.expected_responses[seq_id][0], NodeInfoRequest)
                    and not self.discovery_finished
                ):
                    # Time out for node which is not discovered yet
                    # to speedup the initial discover phase skip retries and mark node as not discovered.
                    _LOGGER.debug(
                        "Skip retry %s to %s to speedup discover process",
                        _request,
                        _mac,
                    )
                    if self.expected_responses[seq_id][1]:
                        self.expected_responses[seq_id][1]()
                else:
                    _LOGGER.info(
                        "Resend %s for %s, retry %s of %s",
                        _request,
                        _mac,
                        str(self.expected_responses[seq_id][2] + 1),
                        str(MESSAGE_RETRY + 1),
                    )
                    self.send(
                        self.expected_responses[seq_id][0],
                        self.expected_responses[seq_id][1],
                        self.expected_responses[seq_id][2] + 1,
                    )
            else:
                _LOGGER.warning(
                    "Drop %s to %s because max retries %s reached",
                    _request,
                    _mac,
                    str(MESSAGE_RETRY + 1),
                )
                # Report node as unavailable for missing NodePingRequest
                if isinstance(self.expected_responses[seq_id][0], NodePingRequest):
                    self.node_state(_mac, False)
                else:
                    _LOGGER.debug(
                        "Do a single ping request to %s to validate if node is reachable",
                        _mac,
                    )
                    self.send(
                        NodePingRequest(self.expected_responses[seq_id][0].mac),
                        None,
                        MESSAGE_RETRY + 1,
                    )
            del self.expected_responses[seq_id]

    def _send_message_loop(self):
        """Daemon to send messages waiting in queue."""
        while self._run_send_message_thread:
            try:
                _prio, _retry, _dt, request_set = self._send_message_queue.get(
                    block=True, timeout=1
                )
            except Empty:
                time.sleep(SLEEP_TIME)
            else:
                # Calc next seq_id based last received ack message
                # if previous seq_id is unknown use fake b"0000"
                seq_id = inc_seq_id(self.last_seq_id)
                self.expected_responses[seq_id] = request_set
                if self.expected_responses[seq_id][2] == 0:
                    _LOGGER.info(
                        "Send %s to %s using seq_id %s",
                        self.expected_responses[seq_id][0].__class__.__name__,
                        self.expected_responses[seq_id][0].mac,
                        str(seq_id),
                    )
                else:
                    _LOGGER.info(
                        "Resend %s to %s using seq_id %s, retry %s",
                        self.expected_responses[seq_id][0].__class__.__name__,
                        self.expected_responses[seq_id][0].mac,
                        str(seq_id),
                        str(self.expected_responses[seq_id][2]),
                    )
                self.expected_responses[seq_id][3] = datetime.now()
                # Send request
                self.connection.send(self.expected_responses[seq_id][0])
                time.sleep(SLEEP_TIME)
                timeout_counter = 0
                # Wait max 1 second for acknowledge response from USB-stick
                while (
                    self.last_seq_id != seq_id
                    and timeout_counter < 10
                    and seq_id != b"0000"
                    and self.last_seq_id is not None
                ):
                    time.sleep(0.1)
                    timeout_counter += 1
                if timeout_counter >= 10 and self._run_send_message_thread:
                    self.resend(seq_id)
        _LOGGER.debug("Send message loop stopped")

    def message_handler(self, message):
        """handle received message from Plugwise Zigbee network."""

        # only save last seq_id and skip special ID's FFFD, FFFE, FFFF
        if self.last_seq_id:
            if int(self.last_seq_id, 16) < int(message.seq_id, 16) < 65533:
                self.last_seq_id = message.seq_id
            elif message.seq_id == b"0000" and self.last_seq_id == b"FFFB":
                self.last_seq_id = b"0000"

        if isinstance(message, NodeAckSmallResponse):
            self._log_status_message(message, message.ack_id)
            self._post_message_action(
                message.seq_id, message.ack_id, message.__class__.__name__
            )
        else:
            if isinstance(message, (NodeAckResponse, NodeAckLargeResponse)):
                self._log_status_message(message, message.ack_id)
            else:
                self._log_status_message(message)
            self.message_processor(message)
            if (
                message.seq_id != b"FFFF"
                and message.seq_id != b"FFFE"
                and message.seq_id != b"FFFD"
            ):
                self._post_message_action(
                    message.seq_id, None, message.__class__.__name__
                )

    def _post_message_action(self, seq_id, ack_response=None, request="unknown"):
        """Execute action if request has been successful.."""
        if seq_id in self.expected_responses:
            if ack_response in (*REQUEST_SUCCESS, None):
                if self.expected_responses[seq_id][1]:
                    _LOGGER.debug(
                        "Execute action %s of request with seq_id %s",
                        self.expected_responses[seq_id][1].__name__,
                        str(seq_id),
                    )
                    try:
                        self.expected_responses[seq_id][1]()
                    # TODO: narrow exception
                    except Exception as err:  # pylint: disable=broad-except
                        _LOGGER.error(
                            "Execution of  %s for request with seq_id %s failed: %s",
                            self.expected_responses[seq_id][1].__name__,
                            str(seq_id),
                            err,
                        )
                del self.expected_responses[seq_id]
            elif ack_response in REQUEST_FAILED:
                self.resend(seq_id)
        else:
            if not self.last_seq_id:
                if b"0000" in self.expected_responses:
                    self.expected_responses[seq_id] = self.expected_responses[b"0000"]
                    del self.expected_responses[b"0000"]
                self.last_seq_id = seq_id
            else:
                _LOGGER.info(
                    "Drop unexpected %s%s using seq_id %s",
                    STATUS_RESPONSES.get(ack_response, "") + " ",
                    request,
                    str(seq_id),
                )

    def _receive_timeout_loop(self):
        """Daemon to time out open requests without any (n)ack response message."""
        while self._run_receive_timeout_thread:
            for seq_id in list(self.expected_responses.keys()):
                if self.expected_responses[seq_id][3] is not None:
                    if self.expected_responses[seq_id][3] < (
                        datetime.now() - timedelta(seconds=MESSAGE_TIME_OUT)
                    ):
                        _mac = "<unknown>"
                        if self.expected_responses[seq_id][0].mac:
                            _mac = self.expected_responses[seq_id][0].mac.decode(
                                UTF8_DECODE
                            )
                        _LOGGER.info(
                            "No response within %s seconds timeout for %s to %s with sequence ID %s",
                            str(MESSAGE_TIME_OUT),
                            self.expected_responses[seq_id][0].__class__.__name__,
                            _mac,
                            str(seq_id),
                        )
                        self.resend(seq_id)
            receive_timeout_checker = 0
            while (
                receive_timeout_checker < MESSAGE_TIME_OUT
                and self._run_receive_timeout_thread
            ):
                time.sleep(1)
                receive_timeout_checker += 1
        _LOGGER.debug("Receive timeout loop stopped")

    def _log_status_message(self, message, status=None):
        """Log status messages.."""
        if status:
            if status in STATUS_RESPONSES:
                _LOGGER.debug(
                    "Received %s %s for request with seq_id %s",
                    STATUS_RESPONSES[status],
                    message.__class__.__name__,
                    str(message.seq_id),
                )
            else:
                if self.expected_responses.get(message.seq_id):
                    _LOGGER.warning(
                        "Received unmanaged (%s) %s in response to %s with seq_id %s",
                        str(status),
                        message.__class__.__name__,
                        str(
                            self.expected_responses[message.seq_id][
                                1
                            ].__class__.__name__
                        ),
                        str(message.seq_id),
                    )
                else:
                    _LOGGER.warning(
                        "Received unmanaged (%s) %s for unknown request with seq_id %s",
                        str(status),
                        message.__class__.__name__,
                        str(message.seq_id),
                    )
        else:
            _LOGGER.info(
                "Received %s from %s with sequence id %s",
                message.__class__.__name__,
                message.mac.decode(UTF8_DECODE),
                str(message.seq_id),
            )

    def disconnect_from_stick(self):
        """Disconnect from stick and raise error if it fails"""
        self._run_send_message_thread = False
        self._run_receive_timeout_thread = False
        self.connection.disconnect()



"""Plugwise Circle+ node object."""
from datetime import datetime
import logging

from ..constants import MAX_TIME_DRIFT, PRIORITY_LOW, UTF8_DECODE
from ..messages.requests import (
    CirclePlusRealTimeClockGetRequest,
    CirclePlusRealTimeClockSetRequest,
    CirclePlusScanRequest,
)
from ..messages.responses import CirclePlusRealTimeClockResponse, CirclePlusScanResponse
from ..nodes.circle import PlugwiseCircle

_LOGGER = logging.getLogger(__name__)


class PlugwiseCirclePlus(PlugwiseCircle):
    """provides interface to the Plugwise Circle+ nodes"""

    def __init__(self, mac, address, message_sender):
        super().__init__(mac, address, message_sender)
        self._plugwise_nodes = {}
        self._scan_response = {}
        self._scan_for_nodes_callback = None
        self._realtime_clock_offset = None
        self.get_real_time_clock(self.sync_realtime_clock)

    def message_for_circle_plus(self, message):
        """
        Process received message
        """
        if isinstance(message, CirclePlusRealTimeClockResponse):
            self._response_realtime_clock(message)
        elif isinstance(message, CirclePlusScanResponse):
            self._process_scan_response(message)
        else:
            _LOGGER.waning(
                "Unsupported message type '%s' received from circle with mac %s",
                str(message.__class__.__name__),
                self.mac,
            )

    def scan_for_nodes(self, callback=None):
        """Scan for registered nodes."""
        self._scan_for_nodes_callback = callback
        for node_address in range(0, 64):
            self.message_sender(CirclePlusScanRequest(self._mac, node_address))
            self._scan_response[node_address] = False

    def _process_scan_response(self, message):
        """Process scan response message."""
        _LOGGER.debug(
            "Process scan response for address %s", message.node_address.value
        )
        if message.node_mac.value != b"FFFFFFFFFFFFFFFF":
            _LOGGER.debug(
                "Linked plugwise node with mac %s found",
                message.node_mac.value.decode(UTF8_DECODE),
            )
            if (
                message.node_mac.value.decode(UTF8_DECODE)
                not in self._plugwise_nodes.keys()
            ):
                self._plugwise_nodes[
                    message.node_mac.value.decode(UTF8_DECODE)
                ] = message.node_address.value
        if self._scan_for_nodes_callback:
            # Check if scan is complete before execute callback
            scan_complete = False
            self._scan_response[message.node_address.value] = True
            for node_address in range(0, 64):
                if not self._scan_response[node_address]:
                    if node_address < message.node_address.value:
                        # Apparently missed response so send new scan request if it's not in queue yet
                        _LOGGER.debug(
                            "Resend missing scan request for address %s",
                            str(node_address),
                        )
                        self.message_sender(
                            CirclePlusScanRequest(self._mac, node_address)
                        )
                    break
                if node_address == 63:
                    scan_complete = True
            if scan_complete and self._scan_for_nodes_callback:
                self._scan_for_nodes_callback(self._plugwise_nodes)
                self._scan_for_nodes_callback = None
                self._plugwise_nodes = {}

    def get_real_time_clock(self, callback=None):
        """get current datetime of internal clock of CirclePlus."""
        self.message_sender(
            CirclePlusRealTimeClockGetRequest(self._mac),
            callback,
            0,
            PRIORITY_LOW,
        )

    def _response_realtime_clock(self, message):
        dt = datetime(
            datetime.now().year,
            datetime.now().month,
            datetime.now().day,
            message.time.value.hour,
            message.time.value.minute,
            message.time.value.second,
        )
        realtime_clock_offset = message.timestamp.replace(microsecond=0) - (
            dt + self.timezone_delta
        )
        if realtime_clock_offset.days == -1:
            self._realtime_clock_offset = realtime_clock_offset.seconds - 86400
        else:
            self._realtime_clock_offset = realtime_clock_offset.seconds
        _LOGGER.debug(
            "Realtime clock of node %s has drifted %s sec",
            self.mac,
            str(self._clock_offset),
        )

    def set_real_time_clock(self, callback=None):
        """set internal clock of CirclePlus."""
        self.message_sender(
            CirclePlusRealTimeClockSetRequest(self._mac, datetime.utcnow()),
            callback,
        )

    def sync_realtime_clock(self, max_drift=0):
        """Sync real time clock of node if time has drifted more than max drifted"""
        if self._realtime_clock_offset is not None:
            if max_drift == 0:
                max_drift = MAX_TIME_DRIFT
            if (self._realtime_clock_offset > max_drift) or (
                self._realtime_clock_offset < -(max_drift)
            ):
                _LOGGER.info(
                    "Reset realtime clock of node %s because time has drifted %s sec",
                    self.mac,
                    str(self._clock_offset),
                )
                self.set_real_time_clock()



"""Base class for serial or socket connections to USB-Stick."""
import logging
import queue
import threading
import time

from ..constants import SLEEP_TIME
from ..messages.requests import NodeRequest

_LOGGER = logging.getLogger(__name__)


class StickConnection:
    """Generic Plugwise stick connection."""

    def __init__(self, port, parser):
        """Initialize StickConnection."""
        self.port = port
        self.parser = parser
        self.run_reader_thread = False
        self.run_writer_thread = False
        self._is_connected = False
        self._writer = None

        self._reader_thread = None
        self._write_queue = None
        self._writer_thread = None

    ################################################
    ###             Open connection              ###
    ################################################

    def connect(self) -> bool:
        """Open the connection."""
        if not self._is_connected:
            self._open_connection()
        return self._is_connected

    ################################################
    ###                   Reader                 ###
    ################################################

    def _reader_start(self, name):
        """Start the reader thread to receive data."""
        self._reader_thread = threading.Thread(None, self._reader_deamon, name, (), {})
        self.run_reader_thread = True
        self._reader_thread.start()

    def _reader_deamon(self):
        """Thread to collect available data from connection."""
        while self.run_reader_thread:
            data = self._read_data()
            if data:
                self.parser(data)
            time.sleep(0.01)
        _LOGGER.debug("Reader daemon stopped")

    ################################################
    ###                 Writer                   ###
    ################################################

    def _writer_start(self, name: str):
        """Start the writer thread to send data."""
        self._write_queue = queue.Queue()
        self._writer_thread = threading.Thread(None, self._writer_daemon, name, (), {})
        self._writer_thread.daemon = True
        self.run_writer_thread = True
        self._writer_thread.start()

    def _writer_daemon(self):
        """Thread to write data from queue to existing connection."""
        while self.run_writer_thread:
            try:
                (message, callback) = self._write_queue.get(block=True, timeout=1)
            except queue.Empty:
                time.sleep(SLEEP_TIME)
            else:
                _LOGGER.debug(
                    "Sending %s to plugwise stick (%s)",
                    message.__class__.__name__,
                    message.serialize(),
                )
                self._write_data(message.serialize())
                time.sleep(SLEEP_TIME)
                if callback:
                    callback()
        _LOGGER.debug("Writer daemon stopped")

    def send(self, message: NodeRequest, callback=None):
        """Add message to write queue."""
        self._write_queue.put_nowait((message, callback))

    ################################################
    ###             Connection state             ###
    ################################################

    def is_connected(self):
        """Return connection state."""
        return self._is_connected

    def read_thread_alive(self):
        """Return state of write thread."""
        return self._reader_thread.isAlive() if self.run_reader_thread else False

    def write_thread_alive(self):
        """Return state of write thread."""
        return self._writer_thread.isAlive() if self.run_writer_thread else False

    ################################################
    ###             Close connection             ###
    ################################################

    def disconnect(self):
        """Close the connection."""
        if self._is_connected:
            self._is_connected = False
            self.run_writer_thread = False
            self.run_reader_thread = False
            max_wait = 5 * SLEEP_TIME
            while self._writer_thread.isAlive():
                time.sleep(SLEEP_TIME)
                max_wait -= SLEEP_TIME
            self._close_connection()



"""Plugwise switch node object."""
import logging

from ..constants import FEATURE_PING, FEATURE_RSSI_IN, FEATURE_RSSI_OUT, FEATURE_SWITCH
from ..messages.responses import NodeSwitchGroupResponse
from ..nodes.sed import NodeSED

_LOGGER = logging.getLogger(__name__)


class PlugwiseSwitch(NodeSED):
    """provides interface to the Plugwise Switch nodes"""

    def __init__(self, mac, address, message_sender):
        super().__init__(mac, address, message_sender)
        self._features = (
            FEATURE_PING["id"],
            FEATURE_RSSI_IN["id"],
            FEATURE_RSSI_OUT["id"],
            FEATURE_SWITCH["id"],
        )
        self._switch_state = False

    @property
    def switch(self) -> bool:
        """Return the last known switch state"""
        return self._switch_state

    def message_for_switch(self, message):
        """
        Process received message
        """
        if isinstance(message, NodeSwitchGroupResponse):
            _LOGGER.debug(
                "Switch group request %s received from %s for group id %s",
                str(message.power_state),
                self.mac,
                str(message.group),
            )
            self._process_switch_group(message)

    def _process_switch_group(self, message):
        """Switch group request from Scan"""
        if message.power_state == 0:
            # turn off => clear motion
            if self._switch_state:
                self._switch_state = False
                self.do_callback(FEATURE_SWITCH["id"])
        elif message.power_state == 1:
            # turn on => motion
            if not self._switch_state:
                self._switch_state = True
                self.do_callback(FEATURE_SWITCH["id"])
        else:
            _LOGGER.debug(
                "Unknown power_state (%s) received from %s",
                str(message.power_state),
                self.mac,
            )

    ## TODO: All functions below can be removed when HA component is changed to use the property values ##
    def get_switch_state(self):
        """Return state of switch"""
        _LOGGER.warning(
            "Function 'get_switch_state' will be removed in future, use the 'switch' property instead !",
        )
        return self._switch_state



"""All known request messages to be send to plugwise devices."""
from ..constants import MESSAGE_FOOTER, MESSAGE_HEADER
from ..messages import PlugwiseMessage
from ..util import (
    DateTime,
    Int,
    LogAddr,
    RealClockDate,
    RealClockTime,
    SInt,
    String,
    Time,
)


class NodeRequest(PlugwiseMessage):
    """Base class for request messages to be send from by USB-Stick."""

    def __init__(self, mac):
        PlugwiseMessage.__init__(self)
        self.args = []
        self.mac = mac


class NodeNetworkInfoRequest(NodeRequest):
    """TODO: PublicNetworkInfoRequest

    No arguments
    """

    ID = b"0001"


class CirclePlusConnectRequest(NodeRequest):
    """
    Request to connect a Circle+ to the Stick

    Response message: CirclePlusConnectResponse
    """

    ID = b"0004"

    # This message has an exceptional format and therefore need to override the serialize method
    def serialize(self):
        # This command has args: byte: key, byte: networkinfo.index, ulong: networkkey = 0
        args = b"00000000000000000000"
        msg = self.ID + args + self.mac
        checksum = self.calculate_checksum(msg)
        return MESSAGE_HEADER + msg + checksum + MESSAGE_FOOTER


class NodeAddRequest(NodeRequest):
    """
    Inform node it is added to the Plugwise Network it to memory of Circle+ node

    Response message: [acknowledge message]
    """

    ID = b"0007"

    def __init__(self, mac, accept: bool):
        super().__init__(mac)
        accept_value = 1 if accept else 0
        self.args.append(Int(accept_value, length=2))

    # This message has an exceptional format (MAC at end of message)
    # and therefore a need to override the serialize method
    def serialize(self):
        args = b"".join(a.serialize() for a in self.args)
        msg = self.ID + args + self.mac
        checksum = self.calculate_checksum(msg)
        return MESSAGE_HEADER + msg + checksum + MESSAGE_FOOTER


class NodeAllowJoiningRequest(NodeRequest):
    """
    Enable or disable receiving joining request of unjoined nodes.
    Circle+ node will respond with an acknowledge message

    Response message: NodeAckLargeResponse
    """

    ID = b"0008"

    def __init__(self, accept: bool):
        super().__init__("")
        # TODO: Make sure that '01' means enable, and '00' disable joining
        val = 1 if accept else 0
        self.args.append(Int(val, length=2))


class NodeResetRequest(NodeRequest):
    """
    TODO: Some kind of reset request

    Response message: ???
    """

    ID = b"0009"

    def __init__(self, mac, moduletype, timeout):
        super().__init__(mac)
        self.args += [
            Int(moduletype, length=2),
            Int(timeout, length=2),
        ]


class StickInitRequest(NodeRequest):
    """
    Initialize USB-Stick

    Response message: StickInitResponse
    """

    ID = b"000A"

    def __init__(self):
        """message for that initializes the Stick"""
        # init is the only request message that doesn't send MAC address
        super().__init__("")


class NodeImagePrepareRequest(NodeRequest):
    """
    TODO: PWEswImagePrepareRequestV1_0

    Response message: TODO:
    """

    ID = b"000B"


class NodePingRequest(NodeRequest):
    """
    Ping node

    Response message: NodePingResponse
    """

    ID = b"000D"


class CirclePowerUsageRequest(NodeRequest):
    """
    Request current power usage

    Response message: CirclePowerUsageResponse
    """

    ID = b"0012"


class CircleClockSetRequest(NodeRequest):
    """
    Set internal clock of node

    Response message: [Acknowledge message]
    """

    ID = b"0016"

    def __init__(self, mac, dt):
        super().__init__(mac)
        passed_days = dt.day - 1
        month_minutes = (passed_days * 24 * 60) + (dt.hour * 60) + dt.minute
        this_date = DateTime(dt.year, dt.month, month_minutes)
        this_time = Time(dt.hour, dt.minute, dt.second)
        day_of_week = Int(dt.weekday(), 2)
        # FIXME: use LogAddr instead
        log_buf_addr = String("FFFFFFFF", 8)
        self.args += [this_date, log_buf_addr, this_time, day_of_week]


class CircleSwitchRelayRequest(NodeRequest):
    """
    switches relay on/off

    Response message: NodeAckLargeResponse
    """

    ID = b"0017"

    def __init__(self, mac, on):
        super().__init__(mac)
        val = 1 if on else 0
        self.args.append(Int(val, length=2))


class CirclePlusScanRequest(NodeRequest):
    """
    Get all linked Circle plugs from Circle+
    a Plugwise network can have 64 devices the node ID value has a range from 0 to 63

    Response message: CirclePlusScanResponse
    """

    ID = b"0018"

    def __init__(self, mac, node_address):
        super().__init__(mac)
        self.args.append(Int(node_address, length=2))
        self.node_address = node_address


class NodeRemoveRequest(NodeRequest):
    """
    Request node to be removed from Plugwise network by
    removing it from memory of Circle+ node.

    Response message: NodeRemoveResponse
    """

    ID = b"001C"

    def __init__(self, mac_circle_plus, mac_to_unjoined):
        super().__init__(mac_circle_plus)
        self.args.append(String(mac_to_unjoined, length=16))


class NodeInfoRequest(NodeRequest):
    """
    Request status info of node

    Response message: NodeInfoResponse
    """

    ID = b"0023"


class CircleCalibrationRequest(NodeRequest):
    """
    Request power calibration settings of node

    Response message: CircleCalibrationResponse
    """

    ID = b"0026"


class CirclePlusRealTimeClockSetRequest(NodeRequest):
    """
    Set real time clock of CirclePlus

    Response message: [Acknowledge message]
    """

    ID = b"0028"

    def __init__(self, mac, dt):
        super().__init__(mac)
        this_time = RealClockTime(dt.hour, dt.minute, dt.second)
        day_of_week = Int(dt.weekday(), 2)
        this_date = RealClockDate(dt.day, dt.month, dt.year)
        self.args += [this_time, day_of_week, this_date]


class CirclePlusRealTimeClockGetRequest(NodeRequest):
    """
    Request current real time clock of CirclePlus

    Response message: CirclePlusRealTimeClockResponse
    """

    ID = b"0029"


class CircleClockGetRequest(NodeRequest):
    """
    Request current internal clock of node

    Response message: CircleClockResponse
    """

    ID = b"003E"


class CircleEnableScheduleRequest(NodeRequest):
    """
    Request to switch Schedule on or off

    Response message: TODO:
    """

    ID = b"0040"

    def __init__(self, mac, on):
        super().__init__(mac)
        val = 1 if on else 0
        self.args.append(Int(val, length=2))
        # the second parameter is always 0x01
        self.args.append(Int(1, length=2))


class NodeAddToGroupRequest(NodeRequest):
    """
    Add node to group

    Response message: TODO:
    """

    ID = b"0045"

    def __init__(self, mac, group_mac, task_id, port_mask):
        super().__init__(mac)
        group_mac_val = String(group_mac, length=16)
        task_id_val = String(task_id, length=16)
        port_mask_val = String(port_mask, length=16)
        self.args += [group_mac_val, task_id_val, port_mask_val]


class NodeRemoveFromGroupRequest(NodeRequest):
    """
    Remove node from group

    Response message: TODO:
    """

    ID = b"0046"

    def __init__(self, mac, group_mac):
        super().__init__(mac)
        group_mac_val = String(group_mac, length=16)
        self.args += [group_mac_val]


class NodeBroadcastGroupSwitchRequest(NodeRequest):
    """
    Broadcast to group to switch

    Response message: TODO:
    """

    ID = b"0047"

    def __init__(self, group_mac, switch_state: bool):
        super().__init__(group_mac)
        val = 1 if switch_state else 0
        self.args.append(Int(val, length=2))


class CirclePowerBufferRequest(NodeRequest):
    """
    Request power usage storaged a given memory address

    Response message: CirclePowerBufferResponse
    """

    ID = b"0048"

    def __init__(self, mac, log_address):
        super().__init__(mac)
        self.args.append(LogAddr(log_address, 8))


class NodeSleepConfigRequest(NodeRequest):
    """
    Configure timers for SED nodes to minimize battery usage

    stay_active             : Duration in seconds the SED will be awake for receiving commands
    sleep_for               : Duration in minutes the SED will be in sleeping mode and not able to respond any command
    maintenance_interval    : Interval in minutes the node will wake up and able to receive commands
    clock_sync              : Enable/disable clock sync
    clock_interval          : Duration in minutes the node synchronize its clock

    Response message: Ack message with SLEEP_SET
    """

    ID = b"0050"

    def __init__(
        self,
        mac,
        stay_active: int,
        maintenance_interval: int,
        sleep_for: int,
        sync_clock: bool,
        clock_interval: int,
    ):
        super().__init__(mac)

        stay_active_val = Int(stay_active, length=2)
        sleep_for_val = Int(sleep_for, length=4)
        maintenance_interval_val = Int(maintenance_interval, length=4)
        val = 1 if sync_clock else 0
        clock_sync_val = Int(val, length=2)
        clock_interval_val = Int(clock_interval, length=4)
        self.args += [
            stay_active_val,
            maintenance_interval_val,
            sleep_for_val,
            clock_sync_val,
            clock_interval_val,
        ]


class NodeSelfRemoveRequest(NodeRequest):
    """
    <command number="0051" vnumber="1.0" implementation="Plugwise.IO.Commands.V20.PWSelfRemovalRequestV1_0">
      <arguments>
        <argument name="macId" length="16"/>
      </arguments>
    </command>

    """

    ID = b"0051"


class NodeMeasureIntervalRequest(NodeRequest):
    """
    Configure the logging interval of power measurement in minutes

    Response message: TODO:
    """

    ID = b"0057"

    def __init__(self, mac, usage, production):
        super().__init__(mac)
        self.args.append(Int(usage, length=4))
        self.args.append(Int(production, length=4))


class NodeClearGroupMacRequest(NodeRequest):
    """
    TODO:

    Response message: ????
    """

    ID = b"0058"

    def __init__(self, mac, taskId):
        super().__init__(mac)
        self.args.append(Int(taskId, length=2))


class CircleSetScheduleValueRequest(NodeRequest):
    """
    Send chunk of On/Off/StandbyKiller Schedule to Circle(+)

    Response message: TODO:
    """

    ID = b"0059"

    def __init__(self, mac, val):
        super().__init__(mac)
        self.args.append(SInt(val, length=4))


class NodeFeaturesRequest(NodeRequest):
    """
    Request feature set node supports

    Response message: NodeFeaturesResponse
    """

    ID = b"005F"


class ScanConfigureRequest(NodeRequest):
    """
    Configure a Scan node

    reset_timer : Delay in minutes when signal is send when no motion is detected
    sensitivity : Sensitivity of Motion sensor (High, Medium, Off)
    light       : Daylight override to only report motion when lightlevel is below calibrated level

    Response message: [Acknowledge message]
    """

    ID = b"0101"

    def __init__(self, mac, reset_timer: int, sensitivity: int, light: bool):
        super().__init__(mac)

        reset_timer_value = Int(reset_timer, length=2)
        # Sensitivity: HIGH(0x14),  MEDIUM(0x1E),  OFF(0xFF)
        sensitivity_value = Int(sensitivity, length=2)
        light_temp = 1 if light else 0
        light_value = Int(light_temp, length=2)
        self.args += [
            sensitivity_value,
            light_value,
            reset_timer_value,
        ]


class ScanLightCalibrateRequest(NodeRequest):
    """
    Calibrate light sensitivity

    Response message: [Acknowledge message]
    """

    ID = b"0102"


class SenseReportIntervalRequest(NodeRequest):
    """
    Sets the Sense temperature and humidity measurement report interval in minutes.
    Based on this interval, periodically a 'SenseReportResponse' message is sent by the Sense node

    Response message: [Acknowledge message]
    """

    ID = b"0102"

    def __init__(self, mac, interval):
        super().__init__(mac)
        self.args.append(Int(interval, length=2))


class CircleInitialRelaisStateRequest(NodeRequest):
    """
    Get or set initial Relais state

    Response message: CircleInitialRelaisStateResponse
    """

    ID = b"0138"

    def __init__(self, mac, configure: bool, relais_state: bool):
        super().__init__(mac)
        set_or_get = Int(1 if configure else 0, length=2)
        relais = Int(1 if relais_state else 0, length=2)
        self.args += [set_or_get, relais]



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Main stick object to control associated plugwise plugs
"""
from datetime import datetime, timedelta
import logging
import sys
import threading
import time

from .constants import (
    ACCEPT_JOIN_REQUESTS,
    CB_JOIN_REQUEST,
    CB_NEW_NODE,
    MESSAGE_TIME_OUT,
    NODE_TYPE_CELSIUS_NR,
    NODE_TYPE_CELSIUS_SED,
    NODE_TYPE_CIRCLE,
    NODE_TYPE_CIRCLE_PLUS,
    NODE_TYPE_SCAN,
    NODE_TYPE_SENSE,
    NODE_TYPE_STEALTH,
    NODE_TYPE_SWITCH,
    PRIORITY_LOW,
    STATE_ACTIONS,
    UTF8_DECODE,
    WATCHDOG_DEAMON,
)
from .controller import StickMessageController
from .exceptions import (
    CirclePlusError,
    NetworkDown,
    PortError,
    StickInitError,
    TimeoutException,
)
from .messages.requests import (
    NodeAddRequest,
    NodeAllowJoiningRequest,
    NodeInfoRequest,
    NodePingRequest,
    NodeRemoveRequest,
    StickInitRequest,
)
from .messages.responses import (
    NodeAckLargeResponse,
    NodeAckResponse,
    NodeInfoResponse,
    NodeJoinAvailableResponse,
    NodeRemoveResponse,
    NodeResponse,
    StickInitResponse,
)
from .nodes import PlugwiseNode
from .nodes.circle import PlugwiseCircle
from .nodes.circle_plus import PlugwiseCirclePlus
from .nodes.scan import PlugwiseScan
from .nodes.sense import PlugwiseSense
from .nodes.stealth import PlugwiseStealth
from .util import validate_mac

_LOGGER = logging.getLogger(__name__)


# TODO: stick should become StickClass be PascalCase style
class stick:
    """Plugwise connection stick."""

    def __init__(self, port, callback=None):
        self.circle_plus_mac = None
        self.init_callback = None
        self.msg_controller = None
        self.scan_callback = None

        self._accept_join_requests = ACCEPT_JOIN_REQUESTS
        self._auto_update_manually = False
        self._auto_update_timer = 0
        self._circle_plus_discovered = False
        self._circle_plus_retries = 0
        self._device_nodes = {}
        self._joined_nodes = 0
        self._mac_stick = None
        self._messages_for_undiscovered_nodes = []
        self._network_id = None
        self._network_online = False
        self._nodes_discovered = None
        self._nodes_not_discovered = {}
        self._nodes_off_line = 0
        self._nodes_to_discover = {}
        self._port = port
        self._run_update_thread = False
        self._run_watchdog = None
        self._stick_callbacks = {}
        self._stick_initialized = False
        self._update_thread = None
        self._watchdog_thread = None

        if callback:
            self.auto_initialize(callback)

    @property
    def devices(self) -> dict:
        """All discovered and supported plugwise devices with the MAC address as their key"""
        return self._device_nodes

    @property
    def joined_nodes(self) -> int:
        """Return total number of nodes registered to Circle+ including Circle+ itself."""
        return self._joined_nodes + 1

    @property
    def mac(self) -> str:
        """Return the MAC address of the USB-Stick."""
        if self._mac_stick:
            return self._mac_stick.decode(UTF8_DECODE)
        return None

    @property
    def network_state(self) -> bool:
        """Return the state of the Plugwise network."""
        return self._network_online

    @property
    def network_id(self) -> int:
        """Return the id of the Plugwise network."""
        return self._network_id

    @property
    def port(self) -> str:
        """Return currently configured port to USB-Stick."""
        return self._port

    @port.setter
    def port(self, port: str):
        """Set port to USB-Stick."""
        if self.msg_controller:
            self.disconnect()
        self._port = port

    def auto_initialize(self, callback=None):
        """Automatic initialization of USB-stick and discovery of all registered nodes."""

        def init_finished():
            if not self._network_online:
                _LOGGER.Error("plugwise Zigbee network down")
            else:
                self.scan(callback)

        if not self.msg_controller:
            self.msg_controller = StickMessageController(
                self.port, self.message_processor, self.node_state_updates
            )
        try:
            self.msg_controller.connect_to_stick()
            self.initialize_stick(init_finished)
        except PortError as err:
            _LOGGER.error("Failed to connect: '%s'", err)
        except StickInitError as err:
            _LOGGER.error("Failed to initialize USBstick: '%s'", err)
        except NetworkDown:
            _LOGGER.error("Failed to communicated: Plugwise Zigbee network")
        except TimeoutException:
            _LOGGER.error("Timeout exception while initializing USBstick")
        except Exception as err:  # pylint: disable=broad-except
            _LOGGER.error("Unknown error : %s", err)

    def connect(self, callback=None):
        """Startup message controller and connect to stick."""
        if not self.msg_controller:
            self.msg_controller = StickMessageController(
                self.port, self.message_processor, self.node_state_updates
            )
        if self.msg_controller.connect_to_stick(callback):
            # update daemon
            self._run_update_thread = False
            self._auto_update_timer = 0
            self._update_thread = threading.Thread(
                None, self._update_loop, "update_thread", (), {}
            )
            self._update_thread.daemon = True

    def initialize_stick(self, callback=None, timeout=MESSAGE_TIME_OUT):
        """Initialize the USB-stick, start watchdog thread and raise an error if this fails."""
        if not self.msg_controller.connection.is_connected():
            raise StickInitError
        _LOGGER.debug("Send init request to Plugwise Zigbee stick")
        self.msg_controller.send(StickInitRequest(), callback)
        time_counter = 0
        while not self._stick_initialized and (time_counter < timeout):
            time_counter += 0.1
            time.sleep(0.1)
        if not self._stick_initialized:
            raise StickInitError
        if not self._network_online:
            raise NetworkDown

    def initialize_circle_plus(self, callback=None, timeout=MESSAGE_TIME_OUT):
        """Initialize connection from USB-Stick to the Circle+/Stealth+ node and raise an error if this fails."""
        if (
            not self.msg_controller.connection.is_connected()
            or not self._stick_initialized
            or not self.circle_plus_mac
        ):
            raise StickInitError
        self.discover_node(self.circle_plus_mac, callback)

        time_counter = 0
        while not self._circle_plus_discovered and (time_counter < timeout):
            time_counter += 0.1
            time.sleep(0.1)
        if not self._circle_plus_discovered:
            raise CirclePlusError

    def disconnect(self):
        """Disconnect from stick and raise error if it fails"""
        self._run_watchdog = False
        self._run_update_thread = False
        self._auto_update_timer = 0
        self.msg_controller.disconnect_from_stick()
        self.msg_controller = None

    def subscribe_stick_callback(self, callback, callback_type):
        """Subscribe callback to execute."""
        if callback_type not in self._stick_callbacks:
            self._stick_callbacks[callback_type] = []
        self._stick_callbacks[callback_type].append(callback)

    def unsubscribe_stick_callback(self, callback, callback_type):
        """Register callback to execute."""
        if callback_type in self._stick_callbacks:
            self._stick_callbacks[callback_type].remove(callback)

    def allow_join_requests(self, enable: bool, accept: bool):
        """
        Enable or disable Plugwise network
        Automatically accept new join request
        """
        self.msg_controller.send(NodeAllowJoiningRequest(enable))
        if enable:
            self._accept_join_requests = accept
        else:
            self._accept_join_requests = False

    def scan(self, callback=None):
        """Scan and try to detect all registered nodes."""
        self.scan_callback = callback
        self.scan_for_registered_nodes()

    def scan_circle_plus(self):
        """Scan the Circle+ memory for registered nodes."""
        if self._device_nodes.get(self.circle_plus_mac):
            _LOGGER.debug("Scan Circle+ for linked nodes...")
            self._device_nodes[self.circle_plus_mac].scan_for_nodes(self.discover_nodes)
        else:
            _LOGGER.error("Circle+ is not discovered yet")

    def scan_for_registered_nodes(self):
        """Discover Circle+ and all registered nodes at Circle+."""
        if self.circle_plus_mac:
            if self._device_nodes.get(self.circle_plus_mac):
                self.scan_circle_plus()
            else:
                _LOGGER.debug("Discover Circle+ at %s", self.circle_plus_mac)
                self.discover_node(self.circle_plus_mac, self.scan_circle_plus)
        else:
            _LOGGER.error(
                "Plugwise stick not properly initialized, Circle+ MAC is missing."
            )

    def discover_nodes(self, nodes_to_discover):
        """Helper to discover all registered nodes."""
        _LOGGER.debug("Scan plugwise network finished")
        self._nodes_discovered = 0
        self._nodes_to_discover = nodes_to_discover
        self._joined_nodes = len(nodes_to_discover)

        # setup timeout for node discovery
        discover_timeout = 10 + (len(nodes_to_discover) * 2) + (MESSAGE_TIME_OUT)
        threading.Timer(discover_timeout, self.scan_timeout_expired).start()
        _LOGGER.debug("Start discovery of linked node types...")
        for mac in nodes_to_discover:
            self.discover_node(mac, self.node_discovered_by_scan)

    def node_discovered_by_scan(self, nodes_off_line=False):
        """Node discovered by initial scan."""
        if nodes_off_line:
            self._nodes_off_line += 1
        self._nodes_discovered += 1
        _LOGGER.debug(
            "Discovered Plugwise node %s (%s off-line) of %s",
            str(len(self._device_nodes)),
            str(self._nodes_off_line),
            str(len(self._nodes_to_discover)),
        )
        if (len(self._device_nodes) - 1 + self._nodes_off_line) >= len(
            self._nodes_to_discover
        ):
            if self._nodes_off_line == 0:
                self._nodes_to_discover = {}
                self._nodes_not_discovered = {}
            else:
                for mac in self._nodes_to_discover:
                    if not self._device_nodes.get(mac):
                        _LOGGER.info(
                            "Failed to discover node type for registered MAC '%s'. This is expected for battery powered nodes, they will be discovered at their first awake",
                            str(mac),
                        )
                    else:
                        if mac in self._nodes_not_discovered:
                            del self._nodes_not_discovered[mac]
            self.msg_controller.discovery_finished = True
            if self.scan_callback:
                self.scan_callback()

    def scan_timeout_expired(self):
        """Timeout for initial scan."""
        if not self.msg_controller.discovery_finished:
            for mac in self._nodes_to_discover:
                if mac not in self._device_nodes.keys():
                    _LOGGER.info(
                        "Failed to discover node type for registered MAC '%s'. This is expected for battery powered nodes, they will be discovered at their first awake",
                        str(mac),
                    )
                else:
                    if mac in self._nodes_not_discovered:
                        del self._nodes_not_discovered[mac]
            if self.scan_callback:
                self.scan_callback()

    def _append_node(self, mac, address, node_type):
        """Add node to list of controllable nodes"""
        _LOGGER.debug(
            "Add new node type (%s) with mac %s",
            str(node_type),
            mac,
        )
        if node_type == NODE_TYPE_CIRCLE_PLUS:
            self._device_nodes[mac] = PlugwiseCirclePlus(
                mac, address, self.msg_controller.send
            )
        elif node_type == NODE_TYPE_CIRCLE:
            self._device_nodes[mac] = PlugwiseCircle(
                mac, address, self.msg_controller.send
            )
        elif node_type == NODE_TYPE_SWITCH:
            self._device_nodes[mac] = None
        elif node_type == NODE_TYPE_SENSE:
            self._device_nodes[mac] = PlugwiseSense(
                mac, address, self.msg_controller.send
            )
        elif node_type == NODE_TYPE_SCAN:
            self._device_nodes[mac] = PlugwiseScan(
                mac, address, self.msg_controller.send
            )
        elif node_type == NODE_TYPE_CELSIUS_SED:
            self._device_nodes[mac] = None
        elif node_type == NODE_TYPE_CELSIUS_NR:
            self._device_nodes[mac] = None
        elif node_type == NODE_TYPE_STEALTH:
            self._device_nodes[mac] = PlugwiseStealth(
                mac, address, self.msg_controller.send
            )
        else:
            _LOGGER.warning("Unsupported node type '%s'", str(node_type))
            self._device_nodes[mac] = None

        # process previous missed messages
        msg_to_process = self._messages_for_undiscovered_nodes[:]
        self._messages_for_undiscovered_nodes = []
        for msg in msg_to_process:
            self.message_processor(msg)

    def node_state_updates(self, mac, state: bool):
        """Update availability state of a node"""
        if self._device_nodes.get(mac):
            if not self._device_nodes[mac].battery_powered:
                self._device_nodes[mac].available = state

    def node_join(self, mac: str, callback=None) -> bool:
        """Accept node to join Plugwise network by register mac in Circle+ memory"""
        if validate_mac(mac):
            self.msg_controller.send(
                NodeAddRequest(bytes(mac, UTF8_DECODE), True), callback
            )
            return True
        _LOGGER.warning("Invalid mac '%s' address, unable to join node manually.", mac)
        return False

    def node_unjoin(self, mac: str, callback=None) -> bool:
        """Remove node from the Plugwise network by deleting mac from the Circle+ memory"""
        if validate_mac(mac):
            self.msg_controller.send(
                NodeRemoveRequest(bytes(self.circle_plus_mac, UTF8_DECODE), mac),
                callback,
            )
            return True

        _LOGGER.warning(
            "Invalid mac '%s' address, unable to unjoin node manually.", mac
        )
        return False

    def _remove_node(self, mac):
        """Remove node from list of controllable nodes."""
        if self._device_nodes.get(mac):
            del self._device_nodes[mac]
        else:
            _LOGGER.warning("Node %s does not exists, unable to remove node.", mac)

    def message_processor(self, message: NodeResponse):
        """Received message from Plugwise network."""
        mac = message.mac.decode(UTF8_DECODE)
        if isinstance(message, (NodeAckLargeResponse, NodeAckResponse)):
            if message.ack_id in STATE_ACTIONS:
                self._pass_message_to_node(message, mac)
        elif isinstance(message, NodeInfoResponse):
            self._process_node_info_response(message, mac)
        elif isinstance(message, StickInitResponse):
            self._process_stick_init_response(message)
        elif isinstance(message, NodeJoinAvailableResponse):
            self._process_node_join_request(message, mac)
        elif isinstance(message, NodeRemoveResponse):
            self._process_node_remove(message)
        else:
            self._pass_message_to_node(message, mac)

    def _process_stick_init_response(self, stick_init_response: StickInitResponse):
        """Process StickInitResponse message."""
        self._mac_stick = stick_init_response.mac
        if stick_init_response.network_is_online.value == 1:
            self._network_online = True
        else:
            self._network_online = False
        # Replace first 2 characters by 00 for mac of circle+ node
        self.circle_plus_mac = "00" + stick_init_response.circle_plus_mac.value[
            2:
        ].decode(UTF8_DECODE)
        self._network_id = stick_init_response.network_id.value
        self._stick_initialized = True
        if not self._run_watchdog:
            self._run_watchdog = True
            self._watchdog_thread = threading.Thread(
                None, self._watchdog_loop, "watchdog_thread", (), {}
            )
            self._watchdog_thread.daemon = True
            self._watchdog_thread.start()

    def _process_node_info_response(self, node_info_response, mac):
        """Process NodeInfoResponse message."""
        if not self._pass_message_to_node(node_info_response, mac, False):
            _LOGGER.debug(
                "Received NodeInfoResponse from currently unknown node with mac %s with sequence id %s",
                mac,
                str(node_info_response.seq_id),
            )
            if node_info_response.node_type.value == NODE_TYPE_CIRCLE_PLUS:
                self._circle_plus_discovered = True
                self._append_node(mac, 0, node_info_response.node_type.value)
                if mac in self._nodes_not_discovered:
                    del self._nodes_not_discovered[mac]
            else:
                if mac in self._nodes_to_discover:
                    _LOGGER.info(
                        "Node with mac %s discovered",
                        mac,
                    )
                    self._append_node(
                        mac,
                        self._nodes_to_discover[mac],
                        node_info_response.node_type.value,
                    )
            self._pass_message_to_node(node_info_response, mac)

    def _process_node_join_request(self, node_join_request, mac):
        """
        Process NodeJoinAvailableResponse message from a node that
        is not part of a plugwise network yet and wants to join
        """
        if self._device_nodes.get(mac):
            _LOGGER.debug(
                "Received node available message for node %s which is already joined.",
                mac,
            )
        else:
            if self._accept_join_requests:
                # Send accept join request
                _LOGGER.info(
                    "Accepting network join request for node with mac %s",
                    mac,
                )
                self.msg_controller.send(NodeAddRequest(node_join_request.mac, True))
                self._nodes_not_discovered[mac] = (None, None)
            else:
                _LOGGER.debug(
                    "New node with mac %s requesting to join Plugwise network, do callback",
                    mac,
                )
                self.do_callback(CB_JOIN_REQUEST, mac)

    def _process_node_remove(self, node_remove_response):
        """
        Process NodeRemoveResponse message with confirmation
        if node is is removed from the Plugwise network.
        """
        unjoined_mac = node_remove_response.node_mac_id.value
        if node_remove_response.status.value == 1:
            if self._device_nodes.get(unjoined_mac):
                del self._device_nodes[unjoined_mac]
                _LOGGER.info(
                    "Received NodeRemoveResponse from node %s it has been unjoined from Plugwise network",
                    unjoined_mac,
                )
            else:
                _LOGGER.debug(
                    "Unknown node with mac %s has been unjoined from Plugwise network",
                    unjoined_mac,
                )
        else:
            _LOGGER.warning(
                "Node with mac %s failed to unjoin from Plugwise network ",
                unjoined_mac,
            )

    def _pass_message_to_node(self, message, mac, discover=True):
        """
        Pass message to node class to take action on message

        Returns True if message has passed onto existing known node
        """
        if self._device_nodes.get(mac):
            self._device_nodes[mac].message_for_node(message)
            return True
        if discover:
            _LOGGER.info(
                "Queue %s from %s because node is not discovered yet.",
                message.__class__.__name__,
                mac,
            )
            self._messages_for_undiscovered_nodes.append(message)
            self.discover_node(mac, self._discover_after_scan, True)
        return False

    def _watchdog_loop(self):
        """
        Main worker loop to watch all other worker threads
        """
        time.sleep(5)
        circle_plus_retry_counter = 0
        while self._run_watchdog:
            # Connection
            if self.msg_controller.connection.is_connected():
                # Connection reader daemon
                if not self.msg_controller.connection.read_thread_alive():
                    _LOGGER.warning("Unexpected halt of connection reader thread")
                # Connection writer daemon
                if not self.msg_controller.connection.write_thread_alive():
                    _LOGGER.warning("Unexpected halt of connection writer thread")
            # receive timeout daemon
            if self.msg_controller._run_receive_timeout_thread:
                if not self.msg_controller._receive_timeout_thread.isAlive():
                    _LOGGER.warning(
                        "Unexpected halt of receive thread, restart thread",
                    )
                    self.msg_controller._receive_timeout_thread = threading.Thread(
                        None,
                        self.msg_controller._receive_timeout_loop,
                        "receive_timeout_thread",
                        (),
                        {},
                    )
                    self.msg_controller._receive_timeout_thread.daemon = True
                    self.msg_controller._receive_timeout_thread.start()
            # send message daemon
            if self.msg_controller._run_send_message_thread:
                if not self.msg_controller._send_message_thread.isAlive():
                    _LOGGER.warning(
                        "Unexpected halt of send thread, restart thread",
                    )
                    self.msg_controller._send_message_thread = threading.Thread(
                        None,
                        self.msg_controller._send_message_loop,
                        "send_messages_thread",
                        (),
                        {},
                    )
                    self.msg_controller._send_message_thread.daemon = True
                    self.msg_controller._send_message_thread.start()
            # Update daemon
            if self._run_update_thread:
                if not self._update_thread.isAlive():
                    _LOGGER.warning(
                        "Unexpected halt of update thread, restart thread",
                    )
                    self._run_update_thread = True
                    self._update_thread = threading.Thread(
                        None, self._update_loop, "update_thread", (), {}
                    )
                    self._update_thread.daemon = True
                    self._update_thread.start()
            # Circle+ discovery
            if not self._circle_plus_discovered:
                # First hour every once an hour
                if self._circle_plus_retries < 60 or circle_plus_retry_counter > 60:
                    _LOGGER.info(
                        "Circle+ not yet discovered, resubmit discovery request",
                    )
                    self.discover_node(self.circle_plus_mac, self.scan)
                    self._circle_plus_retries += 1
                    circle_plus_retry_counter = 0
                circle_plus_retry_counter += 1
            watchdog_loop_checker = 0
            while watchdog_loop_checker < WATCHDOG_DEAMON and self._run_watchdog:
                time.sleep(1)
                watchdog_loop_checker += 1
        _LOGGER.debug("watchdog loop stopped")

    def _update_loop(self):
        """
        When node has not received any message during
        last 2 update polls, reset availability
        """
        self._run_update_thread = True
        _discover_counter = 0
        day_of_month = datetime.now().day
        try:
            while self._run_update_thread:
                for mac in self._device_nodes:
                    if self._device_nodes[mac]:
                        if self._device_nodes[mac].battery_powered:
                            # Check availability state of SED's
                            self._check_availability_of_seds(mac)
                        else:
                            # Do ping request for all non SED's
                            self._device_nodes[mac]._request_ping(None, True)

                        if self._device_nodes[mac].measures_power:
                            # Request current power usage
                            self._device_nodes[mac]._request_power_update()
                            # Sync internal clock of power measure nodes once a day
                            if datetime.now().day != day_of_month:
                                day_of_month = datetime.now().day
                                self._device_nodes[mac].sync_clock()

                # Do a single ping for undiscovered nodes once per 10 update cycles
                if _discover_counter == 10:
                    for mac in self._nodes_not_discovered:
                        self.msg_controller.send(
                            NodePingRequest(bytes(mac, UTF8_DECODE)),
                            None,
                            -1,
                            PRIORITY_LOW,
                        )
                    _discover_counter = 0
                else:
                    _discover_counter += 1

                if self._auto_update_timer and self._run_update_thread:
                    update_loop_checker = 0
                    while (
                        update_loop_checker < self._auto_update_timer
                        and self._run_update_thread
                    ):
                        time.sleep(1)
                        update_loop_checker += 1

        # TODO: narrow exception
        except Exception as err:  # pylint: disable=broad-except
            _exc_type, _exc_obj, exc_tb = sys.exc_info()
            _LOGGER.error(
                "Error at line %s of _update_loop : %s", exc_tb.tb_lineno, err
            )
        _LOGGER.debug("Update loop stopped")

    def auto_update(self, timer=None):
        """Configure auto update polling daemon for power usage and availability state."""
        if timer:
            self._auto_update_timer = timer
            self._auto_update_manually = True
        elif timer == 0:
            self._auto_update_timer = 0
            self._run_update_thread = False
        else:
            # Timer based on a minimum of 5 seconds + 1 second for each node supporting power measurement
            if not self._auto_update_manually:
                count_nodes = 0
                for mac in self._device_nodes:
                    if self._device_nodes[mac].measures_power:
                        count_nodes += 1
                self._auto_update_timer = 5 + (count_nodes * 1)
                _LOGGER.info(
                    "Update interval is (re)set to %s seconds",
                    str(self._auto_update_timer),
                )
        if not self._run_update_thread:
            self._update_thread.start()

    ### Helper functions ###
    def do_callback(self, callback_type, callback_arg=None):
        """Helper to execute registered callbacks for specified callback type."""
        if callback_type in self._stick_callbacks:
            for callback in self._stick_callbacks[callback_type]:
                try:
                    if callback_arg is None:
                        callback()
                    else:
                        callback(callback_arg)
                # TODO: narrow exception
                except Exception as err:  # pylint: disable=broad-except
                    _LOGGER.error("Error while executing callback : %s", err)

    def _check_availability_of_seds(self, mac):
        """Helper to check if SED device is still sending its hartbeat."""
        if self._device_nodes[mac].available:
            if self._device_nodes[mac].last_update < (
                datetime.now()
                - timedelta(minutes=(self._device_nodes[mac].maintenance_interval + 1))
            ):
                _LOGGER.info(
                    "No messages received within (%s minutes) of expected maintenance interval from node %s, mark as unavailable [%s > %s]",
                    str(self._device_nodes[mac].maintenance_interval),
                    mac,
                    str(self._device_nodes[mac].last_update),
                    str(
                        datetime.now()
                        - timedelta(
                            minutes=(self._device_nodes[mac].maintenance_interval + 1)
                        )
                    ),
                )
                self._device_nodes[mac].available = False

    def _discover_after_scan(self):
        """Helper to do callback for new node."""
        node_discovered = None
        for mac in self._nodes_not_discovered:
            if self._device_nodes.get(mac):
                node_discovered = mac
                break
        if node_discovered:
            del self._nodes_not_discovered[node_discovered]
            self.do_callback(CB_NEW_NODE, node_discovered)
            self.auto_update()

    def discover_node(self, mac: str, callback=None, force_discover=False):
        """Helper to try to discovery the node (type) based on mac."""
        if not validate_mac(mac) or self._device_nodes.get(mac):
            return
        if mac not in self._nodes_not_discovered:
            self._nodes_not_discovered[mac] = (
                None,
                None,
            )
            self.msg_controller.send(
                NodeInfoRequest(bytes(mac, UTF8_DECODE)),
                callback,
            )
        else:
            (firstrequest, lastrequest) = self._nodes_not_discovered[mac]
            if not (firstrequest and lastrequest):
                self.msg_controller.send(
                    NodeInfoRequest(bytes(mac, UTF8_DECODE)),
                    callback,
                    0,
                    PRIORITY_LOW,
                )
            elif force_discover:
                self.msg_controller.send(
                    NodeInfoRequest(bytes(mac, UTF8_DECODE)),
                    callback,
                )

    ## TODO: All functions below can be removed when HA component is changed to use the property values ##
    def get_mac_stick(self) -> str:
        """Return mac address of USB-Stick"""
        _LOGGER.warning(
            "Function 'get_mac_stick' will be removed in future, use the 'mac' property instead !",
        )
        if self._mac_stick:
            return self._mac_stick.decode(UTF8_DECODE)
        return None

    def registered_nodes(self) -> int:
        """Return total number of nodes registered to Circle+ including Circle+ itself"""
        _LOGGER.warning(
            "Function 'registered_nodes' will be removed in future, use the 'joined_nodes' property instead !",
        )
        return self._joined_nodes + 1

    def nodes(self) -> list:
        """Return list of mac addresses of discovered and supported plugwise nodes."""
        _LOGGER.warning(
            "Function 'nodes' will be removed in future, use the 'devices' (dict) property instead !",
        )
        return list(
            dict(
                filter(lambda item: item[1] is not None, self._device_nodes.items())
            ).keys()
        )

    def node(self, mac: str) -> PlugwiseNode:
        """Return a specific node object"""
        _LOGGER.warning(
            "Function 'node' will be removed in future, use the 'devices' property (dict) instead !",
        )
        return self._device_nodes.get(mac)



"""Plugwise nodes."""
from datetime import datetime
import logging

from ..constants import (
    FEATURE_AVAILABLE,
    FEATURE_PING,
    FEATURE_RELAY,
    FEATURE_RSSI_IN,
    FEATURE_RSSI_OUT,
    PRIORITY_LOW,
    UTF8_DECODE,
)
from ..messages.requests import NodeFeaturesRequest, NodeInfoRequest, NodePingRequest
from ..messages.responses import (
    NodeFeaturesResponse,
    NodeInfoResponse,
    NodeJoinAckResponse,
    NodePingResponse,
    NodeResponse,
)
from ..util import validate_mac, version_to_model

_LOGGER = logging.getLogger(__name__)


class PlugwiseNode:
    """Base class for a Plugwise node."""

    def __init__(self, mac, address, message_sender):
        mac = mac.upper()
        if not validate_mac(mac):
            _LOGGER.debug(
                "MAC address is in unexpected format: %s",
                str(mac),
            )
        self._mac = bytes(mac, encoding=UTF8_DECODE)
        self.message_sender = message_sender
        self._features = ()
        self._address = address
        self._callbacks = {}
        self._last_update = None
        self._available = False
        self._battery_powered = False
        self._measures_power = False
        self._RSSI_in = None
        self._RSSI_out = None
        self._ping = None
        self._node_type = None
        self._hardware_version = None
        self._firmware_version = None
        self._relay_state = False
        self._last_log_address = None
        self.last_info_message = None
        self._device_features = None

    @property
    def available(self) -> bool:
        """Current network state of plugwise node."""
        return self._available

    @available.setter
    def available(self, state: bool):
        """Set current network availability state of plugwise node."""
        if state:
            if not self._available:
                self._available = True
                _LOGGER.debug(
                    "Mark node %s available",
                    self.mac,
                )
                self.do_callback(FEATURE_AVAILABLE["id"])
        else:
            if self._available:
                self._available = False
                _LOGGER.debug(
                    "Mark node %s unavailable",
                    self.mac,
                )
                self.do_callback(FEATURE_AVAILABLE["id"])

    @property
    def battery_powered(self) -> bool:
        """Return True if node is a SED (battery powered) device."""
        return self._battery_powered

    @property
    def hardware_model(self) -> str:
        """Return hardware model."""
        if self._hardware_version:
            return version_to_model(self._hardware_version)
        return None

    @property
    def hardware_version(self) -> str:
        """Return hardware version."""
        if self._hardware_version is not None:
            return self._hardware_version
        return "Unknown"

    @property
    def features(self) -> tuple:
        """Return the abstracted features supported by this plugwise device."""
        return self._features

    @property
    def firmware_version(self) -> str:
        """Return firmware version."""
        if self._firmware_version is not None:
            return str(self._firmware_version)
        return "Unknown"

    @property
    def last_update(self) -> datetime:
        """Return datetime of last received update."""
        return self._last_update

    @property
    def mac(self) -> str:
        """Return the MAC address in string."""
        return self._mac.decode(UTF8_DECODE)

    @property
    def measures_power(self) -> bool:
        """Return True if node can measure power usage."""
        return self._measures_power

    @property
    def name(self) -> str:
        """Return unique name."""
        return self.hardware_model + " (" + str(self._address) + ")"

    @property
    def ping(self) -> int:
        """Return ping roundtrip in ms."""
        if self._ping is not None:
            return self._ping
        return 0

    @property
    def rssi_in(self) -> int:
        """Return inbound RSSI level."""
        if self._RSSI_in is not None:
            return self._RSSI_in
        return 0

    @property
    def rssi_out(self) -> int:
        """Return outbound RSSI level, based on inbound RSSI level of neighbor node."""
        if self._RSSI_out is not None:
            return self._RSSI_out
        return 0

    def _request_info(self, callback=None):
        """Request info from node."""
        self.message_sender(
            NodeInfoRequest(self._mac),
            callback,
            0,
            PRIORITY_LOW,
        )

    def _request_features(self, callback=None):
        """Request supported features for this node."""
        self.message_sender(
            NodeFeaturesRequest(self._mac),
            callback,
        )

    def _request_ping(self, callback=None, ignore_sensor=True):
        """Ping node."""
        if ignore_sensor or FEATURE_PING["id"] in self._callbacks:
            self.message_sender(
                NodePingRequest(self._mac),
                callback,
            )

    def message_for_node(self, message):
        """Process received message."""
        assert isinstance(message, NodeResponse)
        if message.mac == self._mac:
            if message.timestamp is not None:
                _LOGGER.debug(
                    "Previous update %s of node %s, last message %s",
                    str(self._last_update),
                    self.mac,
                    str(message.timestamp),
                )
                self._last_update = message.timestamp
            if not self._available:
                self.available = True
                self._request_info()
            if isinstance(message, NodePingResponse):
                self._process_ping_response(message)
            elif isinstance(message, NodeInfoResponse):
                self._process_info_response(message)
            elif isinstance(message, NodeFeaturesResponse):
                self._process_features_response(message)
            elif isinstance(message, NodeJoinAckResponse):
                self._process_join_ack_response(message)
            else:
                self.message_for_circle(message)
                self.message_for_sed(message)
        else:
            _LOGGER.debug(
                "Skip message, mac of node (%s) != mac at message (%s)",
                message.mac.decode(UTF8_DECODE),
                self.mac,
            )

    def message_for_circle(self, message):
        """Pass messages to PlugwiseCircle class"""

    def message_for_sed(self, message):
        """Pass messages to NodeSED class"""

    def subscribe_callback(self, callback, sensor) -> bool:
        """Subscribe callback to execute when state change happens."""
        if sensor in self._features:
            if sensor not in self._callbacks:
                self._callbacks[sensor] = []
            self._callbacks[sensor].append(callback)
            return True
        return False

    def unsubscribe_callback(self, callback, sensor):
        """Unsubscribe callback to execute when state change happens."""
        if sensor in self._callbacks:
            self._callbacks[sensor].remove(callback)

    def do_callback(self, sensor):
        """Execute callbacks registered for specified callback type."""
        if sensor in self._callbacks:
            for callback in self._callbacks[sensor]:
                try:
                    callback(None)
                # TODO: narrow exception
                except Exception as err:  # pylint: disable=broad-except
                    _LOGGER.error(
                        "Error while executing all callback : %s",
                        err,
                    )

    def _process_join_ack_response(self, message):
        """Process join acknowledge response message"""
        _LOGGER.info(
            "Node %s has (re)joined plugwise network",
            self.mac,
        )

    def _process_ping_response(self, message):
        """Process ping response message."""
        if self._RSSI_in != message.in_RSSI.value:
            self._RSSI_in = message.in_RSSI.value
            self.do_callback(FEATURE_RSSI_IN["id"])
        if self._RSSI_out != message.out_RSSI.value:
            self._RSSI_out = message.out_RSSI.value
            self.do_callback(FEATURE_RSSI_OUT["id"])
        if self._ping != message.ping_ms.value:
            self._ping = message.ping_ms.value
            self.do_callback(FEATURE_PING["id"])

    def _process_info_response(self, message):
        """Process info response message."""
        _LOGGER.debug("Response info message for node %s", self.mac)
        if message.relay_state.serialize() == b"01":
            if not self._relay_state:
                self._relay_state = True
                self.do_callback(FEATURE_RELAY["id"])
        else:
            if self._relay_state:
                self._relay_state = False
                self.do_callback(FEATURE_RELAY["id"])
        self._hardware_version = message.hw_ver.value.decode(UTF8_DECODE)
        self._firmware_version = message.fw_ver.value
        self._node_type = message.node_type.value
        self.last_info_message = message.timestamp
        if self._last_log_address != message.last_logaddr.value:
            self._last_log_address = message.last_logaddr.value
        _LOGGER.debug("Node type        = %s", self.hardware_model)
        if not self.is_sed:
            _LOGGER.debug("Relay state      = %s", str(self._relay_state))
        _LOGGER.debug("Hardware version = %s", str(self._hardware_version))
        _LOGGER.debug("Firmware version = %s", str(self._firmware_version))

    def _process_features_response(self, message):
        """Process features message."""
        _LOGGER.warning(
            "Node %s supports features %s", self.mac, str(message.features.value)
        )
        self._device_features = message.features.value

    ## TODO: All functions below can be removed when HA component is changed to use the property values ##
    def get_node_type(self) -> str:
        """Return hardware model."""
        _LOGGER.warning(
            "Function 'get_node_type' will be removed in future, use the 'hardware_model' property instead !",
        )
        if self._hardware_version:
            return version_to_model(self._hardware_version)
        return None

    @staticmethod
    def is_sed() -> bool:
        """Return True if node SED (battery powered)."""
        _LOGGER.warning(
            "Function 'is_sed' will be removed in future, use the 'battery_powered' property instead !",
        )
        return False

    def measure_power(self) -> bool:
        """Return True if node can measure power usage."""
        _LOGGER.warning(
            "Function 'measure_power' will be removed in future, use the 'measures_power' property instead !",
        )
        return self._measures_power

    def get_categories(self) -> tuple:
        """Return Home Assistant categories supported by plugwise node."""
        _LOGGER.warning(
            "Function 'get_categories' will be removed in future, use the 'categories' property instead !",
        )
        return self._categories

    def get_sensors(self) -> tuple:
        """Return sensors supported by plugwise node."""
        _LOGGER.warning(
            "Function 'get_sensors' will be removed in future, use the 'sensors' property instead !",
        )
        return self._sensors

    def get_switches(self) -> tuple:
        """Return switches supported by plugwise node."""
        _LOGGER.warning(
            "Function 'get_switches' will be removed in future, use the 'switches' property instead !",
        )
        return self._switches

    def get_available(self) -> bool:
        """Return current network state of plugwise node."""
        _LOGGER.warning(
            "Function 'get_available' will be removed in future, use the 'available' property instead !",
        )
        return self.available

    def set_available(self, state, request_info=False):
        """Set current network availability state of plugwise node."""
        _LOGGER.warning(
            "Function 'set_available' will be removed in future, use the 'available' property instead !",
        )
        self.available = state

    def get_mac(self) -> str:
        """Return mac address."""
        _LOGGER.warning(
            "Function 'get_mac' will be removed in future, use the 'mac' property instead !",
        )
        return self._mac.decode(UTF8_DECODE)

    def get_name(self) -> str:
        """Return unique name."""
        _LOGGER.warning(
            "Function 'get_name' will be removed in future, use the 'name' property instead !",
        )
        return self.self.hardware_model + " (" + str(self._address) + ")"

    def get_hardware_version(self) -> str:
        """Return hardware version."""
        _LOGGER.warning(
            "Function 'get_hardware_version' will be removed in future, use the 'hardware_version' property instead !",
        )
        if self._hardware_version is not None:
            return self._hardware_version
        return "Unknown"

    def get_firmware_version(self) -> str:
        """Return firmware version."""
        _LOGGER.warning(
            "Function 'get_firmware_version' will be removed in future, use the 'firmware_version' property instead !",
        )
        if self._firmware_version is not None:
            return str(self._firmware_version)
        return "Unknown"

    def get_last_update(self) -> datetime:
        """Return  version."""
        _LOGGER.warning(
            "Function 'get_last_update' will be removed in future, use the 'last_update' property instead !",
        )
        return self._last_update

    def get_in_RSSI(self) -> int:
        """Return inbound RSSI level."""
        _LOGGER.warning(
            "Function 'get_in_RSSI' will be removed in future, use the 'rssi_in' property instead !",
        )
        if self._RSSI_in is not None:
            return self._RSSI_in
        return 0

    def get_out_RSSI(self) -> int:
        """Return outbound RSSI level."""
        _LOGGER.warning(
            "Function 'get_out_RSSI' will be removed in future, use the 'rssi_out' property instead !",
        )
        if self._RSSI_out is not None:
            return self._RSSI_out
        return 0

    def get_ping(self) -> int:
        """Return ping roundtrip."""
        _LOGGER.warning(
            "Function 'get_ping' will be removed in future, use the 'ping' property instead !",
        )
        if self._ping is not None:
            return self._ping
        return 0



"""Serial connection."""
import logging

import serial

from ..connections import StickConnection
from ..constants import BAUD_RATE, BYTE_SIZE, STOPBITS
from ..exceptions import PortError

_LOGGER = logging.getLogger(__name__)


class PlugwiseUSBConnection(StickConnection):
    """Simple wrapper around serial module."""

    def __init__(self, port, parser):
        super().__init__(port, parser)
        self._baud = BAUD_RATE
        self._byte_size = BYTE_SIZE
        self._stopbits = STOPBITS
        self._parity = serial.PARITY_NONE

        self._serial = None

    def _open_connection(self):
        """Open serial port."""
        _LOGGER.debug("Open serial port %s", self.port)
        try:
            self._serial = serial.Serial(
                port=self.port,
                baudrate=self._baud,
                bytesize=self._byte_size,
                parity=self._parity,
                stopbits=self._stopbits,
                timeout=1,
            )
        except serial.serialutil.SerialException as err:
            _LOGGER.debug(
                "Failed to connect to serial port %s, %s",
                self.port,
                err,
            )
            raise PortError(err)
        self._is_connected = self._serial.isOpen()
        if self._is_connected:
            self._reader_start("serial_reader_thread")
            self._writer_start("serial_writer_thread")
            _LOGGER.debug("Successfully connected to serial port %s", self.port)
        else:
            _LOGGER.error(
                "Failed to open serial port %s",
                self.port,
            )

    def _close_connection(self):
        """Close serial port."""
        try:
            self._serial.close()
        except serial.serialutil.SerialException as err:
            _LOGGER.debug(
                "Failed to close serial port %s, %s",
                self.port,
                err,
            )
            raise PortError(err)

    def _read_data(self):
        """Read thread."""
        if self._is_connected:
            try:
                serial_data = self._serial.read_all()
            except serial.serialutil.SerialException as err:
                _LOGGER.debug("Error while reading data from serial port : %s", err)
                self._is_connected = False
                raise PortError(err)
            except Exception as err:  # pylint: disable=broad-except
                _LOGGER.debug("Error _read_data : %s", err)
            return serial_data
        return None

    def _write_data(self, data):
        """Write data to serial port."""
        try:
            self._serial.write(data)
        except serial.serialutil.SerialException as err:
            _LOGGER.debug("Error while writing data to serial port : %s", err)
            self._is_connected = False
            raise PortError(err)



"""Plugwise Home Assistant module."""
import asyncio
import logging

import aiohttp

# Dict as class
from munch import Munch

# Version detection
import semver

from .constants import (
    APPLIANCES,
    DEFAULT_PORT,
    DEFAULT_TIMEOUT,
    DEFAULT_USERNAME,
    DOMAIN_OBJECTS,
    LOCATIONS,
    MODULES,
    NOTIFICATIONS,
    RULES,
    SMILES,
    STATUS,
    SWITCH_GROUP_TYPES,
    SYSTEM,
    THERMOSTAT_CLASSES,
)
from .exceptions import ConnectionFailedError, InvalidXMLError, UnsupportedDeviceError
from .helper import SmileHelper

_LOGGER = logging.getLogger(__name__)


class Smile(SmileHelper):
    """Define the Plugwise object."""

    # pylint: disable=too-many-instance-attributes, too-many-public-methods

    def __init__(
        self,
        host,
        password,
        username=DEFAULT_USERNAME,
        port=DEFAULT_PORT,
        timeout=DEFAULT_TIMEOUT,
        websession: aiohttp.ClientSession = None,
    ):
        """Set the constructor for this class."""
        super().__init__()

        if not websession:

            async def _create_session() -> aiohttp.ClientSession:
                return aiohttp.ClientSession()  # pragma: no cover

            loop = asyncio.get_event_loop()
            if loop.is_running():
                self.websession = aiohttp.ClientSession()
            else:
                self.websession = loop.run_until_complete(
                    _create_session()
                )  # pragma: no cover
        else:
            self.websession = websession

        self._auth = aiohttp.BasicAuth(username, password=password)

        self._host = host
        self._port = port
        self._endpoint = f"http://{self._host}:{str(self._port)}"
        self._timeout = timeout

    async def connect(self):
        """Connect to Plugwise device."""
        names = []

        result = await self.request(DOMAIN_OBJECTS)
        dsmrmain = result.find(".//module/protocols/dsmrmain")

        vendor_names = result.findall(".//module/vendor_name")
        for name in vendor_names:
            names.append(name.text)

        if "Plugwise" not in names:
            if dsmrmain is None:  # pragma: no cover
                _LOGGER.error(
                    "Connected but expected text not returned, \
                              we got %s",
                    result,
                )
                raise ConnectionFailedError

        # Determine smile specifics
        await self.smile_detect(result, dsmrmain)

        # Update all endpoints on first connect
        await self.full_update_device()

        return True

    async def smile_detect_legacy(self, result, dsmrmain):
        network = result.find(".//module/protocols/network_router/network")

        # Assume legacy
        self._smile_legacy = True
        # Try if it is an Anna, assuming appliance thermostat
        anna = result.find('.//appliance[type="thermostat"]')
        # Fake insert version assuming Anna
        # couldn't find another way to identify as legacy Anna
        version = "1.8.0"
        model = "smile_thermo"
        if anna is None:
            # P1 legacy:
            if dsmrmain is not None:
                try:
                    status = await self.request(STATUS)
                    version = status.find(".//system/version").text
                    model = status.find(".//system/product").text
                    self.smile_hostname = status.find(".//network/hostname").text
                except InvalidXMLError:  # pragma: no cover
                    # Corner case check
                    raise ConnectionFailedError

            # Stretch:
            elif network is not None:
                try:
                    system = await self.request(SYSTEM)
                    version = system.find(".//gateway/firmware").text
                    model = system.find(".//gateway/product").text
                    self.smile_hostname = system.find(".//gateway/hostname").text
                    self.gateway_id = network.attrib["id"]
                except InvalidXMLError:  # pragma: no cover
                    # Corner case check
                    raise ConnectionFailedError
            else:  # pragma: no cover
                # No cornercase, just end of the line
                _LOGGER.error("Connected but no gateway device information found")
                raise ConnectionFailedError
        return model, version

    async def smile_detect(self, result, dsmrmain):
        """Detect which type of Smile is connected."""
        model = None
        gateway = result.find(".//gateway")

        if gateway is not None:
            model = result.find(".//gateway/vendor_model").text
            version = result.find(".//gateway/firmware_version").text
            if gateway.find("hostname") is not None:
                self.smile_hostname = gateway.find("hostname").text
        else:
            model, version = await self.smile_detect_legacy(result, dsmrmain)

        if model is None or version is None:  # pragma: no cover
            # Corner case check
            _LOGGER.error("Unable to find model or version information")
            raise UnsupportedDeviceError

        ver = semver.VersionInfo.parse(version)
        target_smile = f"{model}_v{ver.major}"

        _LOGGER.debug("Plugwise identified as %s", target_smile)

        if target_smile not in SMILES:
            _LOGGER.error(
                'Your version Smile identified as "%s" seems\
                 unsupported by our plugin, please create an issue\
                 on http://github.com/plugwise/python-plugwise!',
                target_smile,
            )
            raise UnsupportedDeviceError

        self.smile_name = SMILES[target_smile]["friendly_name"]
        self.smile_type = SMILES[target_smile]["type"]
        self.smile_version = (version, ver)

        if "legacy" in SMILES[target_smile]:
            self._smile_legacy = SMILES[target_smile]["legacy"]

        if self.smile_type == "stretch":
            self.stretch_v2 = self.smile_version[1].major == 2
            self.stretch_v3 = self.smile_version[1].major == 3

    async def close_connection(self):
        """Close the Plugwise connection."""
        await self.websession.close()

    async def full_update_device(self):
        """Update all XML data from device."""
        await self.update_domain_objects()
        self._locations = await self.request(LOCATIONS)

        # P1 legacy has no appliances
        if not (self.smile_type == "power" and self._smile_legacy):
            self._appliances = await self.request(APPLIANCES)

        # No need to import modules for P1, no userfull info
        if self.smile_type != "power":
            self._modules = await self.request(MODULES)

    async def update_device(self):
        """Update all XML data from device."""
        await self.update_domain_objects()

        # P1 legacy has no appliances
        if not (self.smile_type == "power" and self._smile_legacy):
            self._appliances = await self.request(APPLIANCES)

    def get_all_devices(self):
        """Determine available devices from inventory."""
        devices = {}
        self.scan_thermostats()

        for appliance, details in self.appl_data.items():
            loc_id = details["location"]
            if loc_id is None:
                details["location"] = self._home_location

            # Override slave thermostat class
            if loc_id in self.thermo_locs:
                if "slaves" in self.thermo_locs[loc_id]:
                    if appliance in self.thermo_locs[loc_id]["slaves"]:
                        details["class"] = "thermo_sensor"

            devices[appliance] = details

        group_data = self.group_switches()
        if group_data is not None:
            devices.update(group_data)

        return devices

    def device_data_switching_group(self, details, device_data):
        """Determine switching groups device data."""
        if details["class"] in SWITCH_GROUP_TYPES:
            counter = 0
            for member in details["members"]:
                appl_data = self.appliance_data(member)
                if appl_data["relay"]:
                    counter += 1

            device_data["relay"] = True
            if counter == 0:
                device_data["relay"] = False

        return device_data

    def device_data_anna(self, dev_id, details, device_data):
        """Determine anna and legacy_anna device data."""
        # Legacy_anna: create Auxiliary heating_state and leave out domestic_hot_water_state
        if "boiler_state" in device_data:
            device_data["heating_state"] = device_data["intended_boiler_state"]
            device_data.pop("boiler_state", None)
            device_data.pop("intended_boiler_state", None)

        # Anna specific
        illuminance = self.object_value("appliance", dev_id, "illuminance")
        if illuminance is not None:
            device_data["illuminance"] = illuminance

        return device_data

    def device_data_adam(self, details, device_data):
        """Determine Adam device data."""
        # Adam: indicate heating_state based on valves being open in case of city-provided heating
        if self.smile_name == "Adam":
            if details["class"] == "gateway":
                if not self.active_device_present and self.heating_valves() is not None:
                    device_data["heating_state"] = True
                    if self.heating_valves() == 0:
                        device_data["heating_state"] = False

        return device_data

    def device_data_climate(self, details, device_data):
        """Determine climate-control device data."""
        # Anna, Lisa, Tom/Floor
        device_data["active_preset"] = self.preset(details["location"])
        device_data["presets"] = self.presets(details["location"])

        avail_schemas, sel_schema, sched_setpoint = self.schemas(details["location"])
        if not self._smile_legacy:
            device_data["schedule_temperature"] = sched_setpoint
        device_data["available_schedules"] = avail_schemas
        device_data["selected_schedule"] = sel_schema
        if self._smile_legacy:
            device_data["last_used"] = "".join(map(str, avail_schemas))
        else:
            device_data["last_used"] = self.last_active_schema(details["location"])

        return device_data

    def get_device_data(self, dev_id):
        """Provide device-data, based on location_id, from APPLIANCES."""
        devices = self.get_all_devices()
        details = devices.get(dev_id)
        device_data = self.appliance_data(dev_id)

        # Generic
        if details["class"] == "gateway" or dev_id == self.gateway_id:
            # Anna: outdoor_temperature only present in domain_objects
            if "outdoor_temperature" not in device_data:
                outdoor_temperature = self.object_value(
                    "location", self._home_location, "outdoor_temperature"
                )
                if outdoor_temperature is not None:
                    device_data["outdoor_temperature"] = outdoor_temperature

            # Try to get P1 data and 2nd outdoor_temperature, when present
            power_data = self.power_data_from_location(details["location"])
            if power_data is not None:
                device_data.update(power_data)

        # Switching groups data
        device_data = self.device_data_switching_group(details, device_data)
        # Specific, not generic Anna data
        device_data = self.device_data_anna(dev_id, details, device_data)
        # Specific, not generic Adam data
        device_data = self.device_data_adam(details, device_data)
        # Unless thermostat based, no need to walk presets
        if details["class"] not in THERMOSTAT_CLASSES:
            return device_data

        # Climate based data (presets, temperatures etc)
        device_data = self.device_data_climate(details, device_data)

        return device_data

    def single_master_thermostat(self):
        """Determine if there is a single master thermostat in the setup."""
        if self.smile_type != "thermostat":
            self.thermo_locs = self.match_locations()
            return None

        count = 0
        self.scan_thermostats()
        for dummy, data in self.thermo_locs.items():
            if "master_prio" in data:
                if data.get("master_prio") > 0:
                    count += 1

        if count == 1:
            return True
        return False

    async def set_schedule_state(self, loc_id, name, state):
        """
        Set the schedule, with the given name, connected to a location.

        Determined from - DOMAIN_OBJECTS.
        """
        if self._smile_legacy:
            return await self.set_schedule_state_legacy(name, state)

        schema_rule_ids = self.rule_ids_by_name(str(name), loc_id)
        if schema_rule_ids == {} or schema_rule_ids is None:
            return False

        for schema_rule_id, location_id in schema_rule_ids.items():
            template_id = None
            if location_id == loc_id:
                state = str(state)
                locator = f'.//*[@id="{schema_rule_id}"]/template'
                for rule in self._domain_objects.findall(locator):
                    template_id = rule.attrib["id"]

                uri = f"{RULES};id={schema_rule_id}"
                data = (
                    "<rules><rule"
                    f' id="{schema_rule_id}"><name><![CDATA[{name}]]></name><template'
                    f' id="{template_id}"/><active>{state}</active></rule></rules>'
                )

                await self.request(uri, method="put", data=data)

        return True

    async def set_preset(self, loc_id, preset):
        """Set the given location-preset on the relevant thermostat - from LOCATIONS."""
        if self._smile_legacy:
            return await self.set_preset_legacy(preset)

        current_location = self._locations.find(f'location[@id="{loc_id}"]')
        location_name = current_location.find("name").text
        location_type = current_location.find("type").text

        if preset not in self.presets(loc_id):
            return False

        uri = f"{LOCATIONS};id={loc_id}"
        data = (
            "<locations><location"
            f' id="{loc_id}"><name>{location_name}</name><type>{location_type}'
            f"</type><preset>{preset}</preset></location></locations>"
        )

        await self.request(uri, method="put", data=data)
        return True

    async def set_temperature(self, loc_id, temperature):
        """Send temperature-set request to the locations thermostat."""
        temperature = str(temperature)
        uri = self.temperature_uri(loc_id)
        data = (
            "<thermostat_functionality><setpoint>"
            f"{temperature}</setpoint></thermostat_functionality>"
        )

        await self.request(uri, method="put", data=data)
        return True

    async def set_groupswitch_member_state(self, members, state, switch):
        """Switch the Switch within a group of members off/on."""
        for member in members:
            locator = f'appliance[@id="{member}"]/{switch.actuator}/{switch.func_type}'
            switch_id = self._appliances.find(locator).attrib["id"]
            uri = f"{APPLIANCES};id={member}/{switch.device};id={switch_id}"
            if self.stretch_v2:
                uri = f"{APPLIANCES};id={member}/{switch.device}"
            state = str(state)
            data = f"<{switch.func_type}><state>{state}</state></{switch.func_type}>"

            await self.request(uri, method="put", data=data)

        return True

    async def set_switch_state(self, appl_id, members, model, state):
        """Switch the Switch off/on."""
        switch = Munch()
        switch.actuator = "actuator_functionalities"
        switch.device = "relay"
        switch.func_type = "relay_functionality"
        switch.func = "state"
        if model == "dhw_cm_switch":
            switch.device = "toggle"
            switch.func_type = "toggle_functionality"

        if model == "lock":
            switch.func = "lock"
            state = "false" if state == "off" else "true"

        if self.stretch_v2:
            switch.actuator = "actuators"
            switch.func_type = "relay"

        if members is not None:
            return await self.set_groupswitch_member_state(members, state, switch)

        locator = f'appliance[@id="{appl_id}"]/{switch.actuator}/{switch.func_type}'
        switch_id = self._appliances.find(locator).attrib["id"]
        uri = f"{APPLIANCES};id={appl_id}/{switch.device};id={switch_id}"
        if self.stretch_v2:
            uri = f"{APPLIANCES};id={appl_id}/{switch.device}"
        data = f"<{switch.func_type}><{switch.func}>{state}</{switch.func}></{switch.func_type}>"

        if model == "relay":
            locator = (
                f'appliance[@id="{appl_id}"]/{switch.actuator}/{switch.func_type}/lock'
            )
            lock_state = self._appliances.find(locator).text
            print("Lock state: ", lock_state)
            # Don't bother switching a relay when the corresponding lock-state is true
            if lock_state == "true":
                return False
            await self.request(uri, method="put", data=data)
            return True

        await self.request(uri, method="put", data=data)
        return True

    async def set_preset_legacy(self, preset):
        """Set the given preset on the thermostat - from DOMAIN_OBJECTS."""
        locator = f'rule/directives/when/then[@icon="{preset}"].../.../...'
        rule = self._domain_objects.find(locator)
        if rule is None:
            return False

        uri = f"{RULES}"
        data = f'<rules><rule id="{rule.attrib["id"]}"><active>true</active></rule></rules>'

        await self.request(uri, method="put", data=data)
        return True

    async def set_schedule_state_legacy(self, name, state):
        """Send a set request to the schema with the given name."""
        schema_rule_id = None
        for rule in self._domain_objects.findall("rule"):
            if rule.find("name").text == name:
                schema_rule_id = rule.attrib["id"]

        if schema_rule_id is None:
            return False

        template_id = None
        state = str(state)
        locator = f'.//*[@id="{schema_rule_id}"]/template'
        for rule in self._domain_objects.findall(locator):
            template_id = rule.attrib["id"]

        uri = f"{RULES};id={schema_rule_id}"
        data = (
            "<rules><rule"
            f' id="{schema_rule_id}"><name><![CDATA[{name}]]></name><template'
            f' id="{template_id}" /><active>{state}</active></rule></rules>'
        )

        await self.request(uri, method="put", data=data)
        return True

    async def delete_notification(self):
        """Send a set request to the schema with the given name."""
        uri = f"{NOTIFICATIONS}"

        await self.request(uri, method="delete")
        return True



"""Plugwise messages."""

from ..constants import MESSAGE_FOOTER, MESSAGE_HEADER, UTF8_DECODE
from ..util import crc_fun


class PlugwiseMessage:
    """Plugwise message base."""

    def __init__(self):
        self.mac = ""
        self.checksum = None
        self.args = []

    def serialize(self):
        """Return message in a serialized format that can be sent out on wire."""
        _args = b"".join(a.serialize() for a in self.args)
        msg = self.ID
        if self.mac != "":
            msg += self.mac
        msg += _args
        self.checksum = self.calculate_checksum(msg)
        return MESSAGE_HEADER + msg + self.checksum + MESSAGE_FOOTER

    @staticmethod
    def calculate_checksum(something):
        """Calculate crc checksum."""
        return bytes("%04X" % crc_fun(something), UTF8_DECODE)



"""
Data parser for USB-Stick

The parser will:
- buffer receiving data
- filter out received zigbee routing data
- collect message data by detecting header and footer
- detect message type based on message ID or fixed sequence ID
- validate received data on checksum
- decode collected data into a response message instance
- pass over received messages to message_processor (controller.py)

"""

import logging

from .constants import MESSAGE_FOOTER, MESSAGE_HEADER
from .exceptions import (
    InvalidMessageChecksum,
    InvalidMessageFooter,
    InvalidMessageHeader,
    InvalidMessageLength,
)
from .messages.responses import get_message_response

_LOGGER = logging.getLogger(__name__)


class PlugwiseParser:
    """Transform Plugwise message from wire format to response message object."""

    def __init__(self, message_processor):
        self.message_processor = message_processor
        self._buffer = bytes([])
        self._parsing = False
        self._message = None

    def feed(self, data):
        """
        Add new incoming data to buffer and try to process
        """
        _LOGGER.debug("Feed data: %s", str(data))
        self._buffer += data
        if len(self._buffer) >= 8:
            if not self._parsing:
                self.parse_data()

    def next_message(self, message):
        """
        Process next packet if present
        """
        try:
            self.message_processor(message)
        # TODO: narrow exception
        except Exception as err:  # pylint: disable=broad-except
            _LOGGER.error(
                "Error while processing %s : %s",
                self._message.__class__.__name__,
                err,
            )
            _LOGGER.error(err, exc_info=True)

    def parse_data(self):
        """
        Process next set of packet data
        """
        _LOGGER.debug("Parse data: %s ", str(self._buffer))
        if not self._parsing:
            self._parsing = True

            # Lookup header of message in buffer
            _LOGGER.debug(
                "Lookup message header (%s) in (%s)",
                str(MESSAGE_HEADER),
                str(self._buffer),
            )
            header_index = self._buffer.find(MESSAGE_HEADER)
            if header_index == -1:
                _LOGGER.debug("No valid message header found yet")
            else:
                _LOGGER.debug(
                    "Valid message header found at index %s", str(header_index)
                )
                self._buffer = self._buffer[header_index:]

                # Header available, lookup footer of message in buffer
                _LOGGER.debug(
                    "Lookup message footer (%s) in (%s)",
                    str(MESSAGE_FOOTER),
                    str(self._buffer),
                )
                footer_index = self._buffer.find(MESSAGE_FOOTER)
                if footer_index == -1:
                    _LOGGER.debug("No valid message footer found yet")
                else:
                    _LOGGER.debug(
                        "Valid message footer found at index %s", str(footer_index)
                    )
                    self._message = get_message_response(
                        self._buffer[4:8], footer_index, self._buffer[8:12]
                    )
                    if self._message:
                        try:
                            self._message.deserialize(self._buffer[: footer_index + 2])
                        except (
                            InvalidMessageChecksum,
                            InvalidMessageFooter,
                            InvalidMessageHeader,
                            InvalidMessageLength,
                        ) as err:
                            _LOGGER.warning(err)
                        # TODO: narrow exception
                        except Exception as err:  # pylint: disable=broad-except
                            _LOGGER.error(
                                "Failed to parse %s message (%s)",
                                self._message.__class__.__name__,
                                str(self._buffer[: footer_index + 2]),
                            )
                            _LOGGER.error(err)
                        else:
                            # Submit message
                            self.next_message(self._message)
                        # Parse remaining buffer
                        self.reset_parser(self._buffer[footer_index + 2 :])
                    else:
                        # skip this message, so remove header from buffer
                        _LOGGER.error(
                            "Skip unknown message %s",
                            str(self._buffer[: footer_index + 2]),
                        )
                        self.reset_parser(self._buffer[6:])
            self._parsing = False
        else:
            _LOGGER.debug("Skip parsing session")

    def reset_parser(self, new_buffer=bytes([])):
        _LOGGER.debug("Reset parser : %s", new_buffer)
        if new_buffer == b"\x83":
            # Skip additional byte sometimes appended after footer
            self._buffer = bytes([])
        else:
            self._buffer = new_buffer
        self._message = None
        self._parsing = False
        if len(self._buffer) > 0:
            self.parse_data()



"""Plugwise Sense node object."""
import logging

from ..constants import (
    FEATURE_HUMIDITY,
    FEATURE_PING,
    FEATURE_RSSI_IN,
    FEATURE_RSSI_OUT,
    FEATURE_TEMPERATURE,
    SENSE_HUMIDITY_MULTIPLIER,
    SENSE_HUMIDITY_OFFSET,
    SENSE_TEMPERATURE_MULTIPLIER,
    SENSE_TEMPERATURE_OFFSET,
)
from ..messages.responses import SenseReportResponse
from ..nodes.sed import NodeSED

_LOGGER = logging.getLogger(__name__)


class PlugwiseSense(NodeSED):
    """provides interface to the Plugwise Sense nodes"""

    def __init__(self, mac, address, message_sender):
        super().__init__(mac, address, message_sender)
        self._features = (
            FEATURE_HUMIDITY["id"],
            FEATURE_PING["id"],
            FEATURE_RSSI_IN["id"],
            FEATURE_RSSI_OUT["id"],
            FEATURE_TEMPERATURE["id"],
        )
        self._temperature = None
        self._humidity = None

    @property
    def humidity(self) -> int:
        """Return the current humidity."""
        return self._humidity

    @property
    def temperature(self) -> int:
        """Return the current temperature."""
        return self._temperature

    def message_for_sense(self, message):
        """
        Process received message
        """
        if isinstance(message, SenseReportResponse):
            self._process_sense_report(message)
        else:
            _LOGGER.info(
                "Unsupported message %s received from %s",
                message.__class__.__name__,
                self.mac,
            )

    def _process_sense_report(self, message):
        """process sense report message to extract current temperature and humidity values."""
        if message.temperature.value != 65535:
            new_temperature = int(
                SENSE_TEMPERATURE_MULTIPLIER * (message.temperature.value / 65536)
                - SENSE_TEMPERATURE_OFFSET
            )
            if self._temperature != new_temperature:
                self._temperature = new_temperature
                _LOGGER.debug(
                    "Sense report received from %s with new temperature level of %s",
                    self.mac,
                    str(self._temperature),
                )
                self.do_callback(FEATURE_TEMPERATURE["id"])
        if message.humidity.value != 65535:
            new_humidity = int(
                SENSE_HUMIDITY_MULTIPLIER * (message.humidity.value / 65536)
                - SENSE_HUMIDITY_OFFSET
            )
            if self._humidity != new_humidity:
                self._humidity = new_humidity
                _LOGGER.debug(
                    "Sense report received from %s with new humidity level of %s",
                    self.mac,
                    str(self._humidity),
                )
                self.do_callback(FEATURE_HUMIDITY["id"])

    ## TODO: All functions below can be removed when HA component is changed to use the property values ##
    def get_temperature(self) -> int:
        """Return the current temperature."""
        _LOGGER.warning(
            "Function 'get_temperature' will be removed in future, use the 'temperature' property instead !",
        )
        return self._temperature

    def get_humidity(self) -> int:
        """Return the current humidity."""
        _LOGGER.warning(
            "Function 'get_humidity' will be removed in future, use the 'humidity' property instead !",
        )
        return self._humidity



"""All known response messages to be received from plugwise devices."""
from datetime import datetime

from ..constants import MESSAGE_FOOTER, MESSAGE_HEADER, MESSAGE_LARGE, MESSAGE_SMALL
from ..exceptions import (
    InvalidMessageChecksum,
    InvalidMessageFooter,
    InvalidMessageHeader,
    InvalidMessageLength,
)
from ..messages import PlugwiseMessage
from ..util import (
    DateTime,
    Float,
    Int,
    LogAddr,
    RealClockDate,
    RealClockTime,
    String,
    Time,
    UnixTimestamp,
)


class NodeResponse(PlugwiseMessage):
    """
    Base class for response messages received by USB-Stick.
    """

    def __init__(self, format_size=None):
        super().__init__()
        self.format_size = format_size
        self.params = []
        self.timestamp = None
        self.seq_id = None
        self.msg_id = None
        self.ack_id = None
        if self.format_size == MESSAGE_SMALL:
            self.len_correction = -12
        elif self.format_size == MESSAGE_LARGE:
            self.len_correction = 4
        else:
            self.len_correction = 0

    def deserialize(self, response):
        self.timestamp = datetime.now()
        if response[:4] != MESSAGE_HEADER:
            raise InvalidMessageHeader(
                f"Invalid message header {str(response[:4])} for {self.__class__.__name__}"
            )
        if response[-2:] != MESSAGE_FOOTER:
            raise InvalidMessageFooter(
                f"Invalid message footer {str(response[-2:])} for {self.__class__.__name__}"
            )
        _calculated_checksum = self.calculate_checksum(response[4:-6])
        _message_checksum = response[-6:-2]
        if _calculated_checksum != _message_checksum:
            raise InvalidMessageChecksum(
                f"Invalid checksum for {self.__class__.__name__}, expected {str(_calculated_checksum)} got {str(_message_checksum)}",
            )
        if len(response) != len(self):
            raise InvalidMessageLength(
                f"Invalid message length received for {self.__class__.__name__}, expected {str(len(self))} bytes got {str(len(response))}"
            )

        self.msg_id = response[4:8]
        self.seq_id = response[8:12]
        response = response[12:]
        if self.format_size == MESSAGE_SMALL or self.format_size == MESSAGE_LARGE:
            self.ack_id = response[:4]
            response = response[4:]
        if self.format_size != MESSAGE_SMALL:
            self.mac = response[:16]
            response = response[16:]
        response = self._parse_params(response)

        _args = b"".join(a.serialize() for a in self.args)
        msg = self.ID
        if self.mac != "":
            msg += self.mac
        msg += _args

    def _parse_params(self, response):
        for param in self.params:
            my_val = response[: len(param)]
            param.deserialize(my_val)
            response = response[len(my_val) :]
        return response

    def __len__(self):
        arglen = sum(len(x) for x in self.params)
        return 34 + arglen + self.len_correction


class NodeAckSmallResponse(NodeResponse):
    """
    Acknowledge message without source MAC

    Response to: Any message
    """

    ID = b"0000"

    def __init__(self):
        super().__init__(MESSAGE_SMALL)


class NodeAckLargeResponse(NodeResponse):
    """
    Acknowledge message with source MAC

    Response to: Any message
    """

    ID = b"0000"

    def __init__(self):
        super().__init__(MESSAGE_LARGE)


class CirclePlusQueryResponse(NodeResponse):
    """
    TODO:

    Response to : ???
    """

    ID = b"0002"

    def __init__(self):
        super().__init__()
        self.channel = String(None, length=2)
        self.source_mac_id = String(None, length=16)
        self.extended_pan_id = String(None, length=16)
        self.unique_network_id = String(None, length=16)
        self.new_node_mac_id = String(None, length=16)
        self.pan_id = String(None, length=4)
        self.idx = Int(0, length=2)
        self.params += [
            self.channel,
            self.source_mac_id,
            self.extended_pan_id,
            self.unique_network_id,
            self.new_node_mac_id,
            self.pan_id,
            self.idx,
        ]

    def __len__(self):
        arglen = sum(len(x) for x in self.params)
        return 18 + arglen

    def deserialize(self, response):
        super().deserialize(response)
        # Clear first two characters of mac ID, as they contain part of the short PAN-ID
        self.new_node_mac_id.value = b"00" + self.new_node_mac_id.value[2:]


class CirclePlusQueryEndResponse(NodeResponse):
    """
    TODO:
        PWAckReplyV1_0
        <argument name="code" length="2"/>

    Response to : ???
    """

    ID = b"0003"

    def __init__(self):
        super().__init__()
        self.status = Int(0, 4)
        self.params += [self.status]

    def __len__(self):
        arglen = sum(len(x) for x in self.params)
        return 18 + arglen


class CirclePlusConnectResponse(NodeResponse):
    """
    CirclePlus connected to the network

    Response to : CirclePlusConnectRequest
    """

    ID = b"0005"

    def __init__(self):
        super().__init__()
        self.existing = Int(0, 2)
        self.allowed = Int(0, 2)
        self.params += [self.existing, self.allowed]

    def __len__(self):
        arglen = sum(len(x) for x in self.params)
        return 18 + arglen


class NodeJoinAvailableResponse(NodeResponse):
    """
    Message from an unjoined node to notify it is available to join a plugwise network

    Response to : <nothing>
    """

    ID = b"0006"


class StickInitResponse(NodeResponse):
    """
    Returns the configuration and status of the USB-Stick

    Optional:
    - circle_plus_mac
    - network_id

    <argument name="upYesNo" length="2"/>
    <argument name="extendedPanId" length="16" optional="1"/>
    <argument name="panId" length="4" optional="1"/>

    Response to: StickInitRequest
    """

    ID = b"0011"

    def __init__(self):
        super().__init__()
        self.unknown1 = Int(0, length=2)
        self.network_is_online = Int(0, length=2)
        self.circle_plus_mac = String(None, length=16)
        self.network_id = Int(0, 4, False)
        self.unknown2 = Int(0, length=2)
        self.params += [
            self.unknown1,
            self.network_is_online,
            self.circle_plus_mac,
            self.network_id,
            self.unknown2,
        ]


class NodePingResponse(NodeResponse):
    """
    Ping response from node

    - incomingLastHopRssiTarget
    - lastHopRssiSource
    - timediffInMs

    Response to : NodePingRequest
    """

    ID = b"000E"

    def __init__(self):
        super().__init__()
        self.in_RSSI = Int(0, length=2)
        self.out_RSSI = Int(0, length=2)
        self.ping_ms = Int(0, 4, False)
        self.params += [
            self.in_RSSI,
            self.out_RSSI,
            self.ping_ms,
        ]


class CirclePowerUsageResponse(NodeResponse):
    """
    Returns power usage as impulse counters for several different timeframes

    Response to : CirclePowerUsageRequest
    """

    ID = b"0013"

    def __init__(self):
        super().__init__()
        self.pulse_1s = Int(0, 4)
        self.pulse_8s = Int(0, 4)
        self.pulse_hour_consumed = Int(0, 8)
        self.pulse_hour_produced = Int(0, 8)
        self.nanosecond_offset = Int(0, 4)
        self.params += [
            self.pulse_1s,
            self.pulse_8s,
            self.pulse_hour_consumed,
            self.pulse_hour_produced,
            self.nanosecond_offset,
        ]


class CirclePlusScanResponse(NodeResponse):
    """
    Returns the MAC of a registered node at the specified memory address

    Response to: CirclePlusScanRequest
    """

    ID = b"0019"

    def __init__(self):
        super().__init__()
        self.node_mac = String(None, length=16)
        self.node_address = Int(0, 2, False)
        self.params += [self.node_mac, self.node_address]


class NodeRemoveResponse(NodeResponse):
    """
    Returns conformation (or not) if node is removed from the Plugwise network
    by having it removed from the memory of the Circle+

    Response to: NodeRemoveRequest
    """

    ID = b"001D"

    def __init__(self):
        super().__init__()
        self.node_mac_id = String(None, length=16)
        self.status = Int(0, 2)
        self.params += [self.node_mac_id, self.status]


class NodeInfoResponse(NodeResponse):
    """
    Returns the status information of Node

    Response to: NodeInfoRequest
    """

    ID = b"0024"

    def __init__(self):
        super().__init__()
        self.datetime = DateTime()
        self.last_logaddr = LogAddr(0, length=8)
        self.relay_state = Int(0, length=2)
        self.hz = Int(0, length=2)
        self.hw_ver = String(None, length=12)
        self.fw_ver = UnixTimestamp(0)
        self.node_type = Int(0, length=2)
        self.params += [
            self.datetime,
            self.last_logaddr,
            self.relay_state,
            self.hz,
            self.hw_ver,
            self.fw_ver,
            self.node_type,
        ]


class CircleCalibrationResponse(NodeResponse):
    """
    returns the calibration settings of node

    Response to: CircleCalibrationRequest
    """

    ID = b"0027"

    def __init__(self):
        super().__init__()
        self.gain_a = Float(0, 8)
        self.gain_b = Float(0, 8)
        self.off_tot = Float(0, 8)
        self.off_noise = Float(0, 8)
        self.params += [self.gain_a, self.gain_b, self.off_tot, self.off_noise]


class CirclePlusRealTimeClockResponse(NodeResponse):
    """
    returns the real time clock of CirclePlus node

    Response to: CirclePlusRealTimeClockGetRequest
    """

    ID = b"003A"

    def __init__(self):
        super().__init__()

        self.time = RealClockTime()
        self.day_of_week = Int(0, 2, False)
        self.date = RealClockDate()
        self.params += [self.time, self.day_of_week, self.date]


class CircleClockResponse(NodeResponse):
    """
    Returns the current internal clock of Node

    Response to: CircleClockGetRequest
    """

    ID = b"003F"

    def __init__(self):
        super().__init__()
        self.time = Time()
        self.day_of_week = Int(0, 2, False)
        self.unknown = Int(0, 2)
        self.unknown2 = Int(0, 4)
        self.params += [self.time, self.day_of_week, self.unknown, self.unknown2]


class CirclePowerBufferResponse(NodeResponse):
    """
    returns information about historical power usage
    each response contains 4 log buffers and each log buffer contains data for 1 hour

    Response to: CirclePowerBufferRequest
    """

    ID = b"0049"

    def __init__(self):
        super().__init__()
        self.logdate1 = DateTime()
        self.pulses1 = Int(0, 8)
        self.logdate2 = DateTime()
        self.pulses2 = Int(0, 8)
        self.logdate3 = DateTime()
        self.pulses3 = Int(0, 8)
        self.logdate4 = DateTime()
        self.pulses4 = Int(0, 8)
        self.logaddr = LogAddr(0, length=8)
        self.params += [
            self.logdate1,
            self.pulses1,
            self.logdate2,
            self.pulses2,
            self.logdate3,
            self.pulses3,
            self.logdate4,
            self.pulses4,
            self.logaddr,
        ]


class NodeAwakeResponse(NodeResponse):
    """
    A sleeping end device (SED: Scan, Sense, Switch) sends
    this message to announce that is awake. Awake types:
    - 0 : The SED joins the network for maintenance
    - 1 : The SED joins a network for the first time
    - 2 : The SED joins a network it has already joined, e.g. after reinserting a battery
    - 3 : When a SED switches a device group or when reporting values such as temperature/humidity
    - 4 : TODO: Unknown
    - 5 : A human pressed the button on a SED to wake it up

    Response to: <nothing>
    """

    ID = b"004F"

    def __init__(self):
        super().__init__()
        self.awake_type = Int(0, 2, False)
        self.params += [self.awake_type]


class NodeSwitchGroupResponse(NodeResponse):
    """
    A sleeping end device (SED: Scan, Sense, Switch) sends
    this message to switch groups on/off when the configured
    switching conditions have been met.

    Response to: <nothing>
    """

    ID = b"0056"

    def __init__(self):
        super().__init__()
        self.group = Int(0, 2, False)
        self.power_state = Int(0, length=2)
        self.params += [
            self.group,
            self.power_state,
        ]


class NodeFeaturesResponse(NodeResponse):
    """
    Returns supported features of node
    TODO: FeatureBitmask

    Response to: NodeFeaturesRequest
    """

    ID = b"0060"

    def __init__(self):
        super().__init__()
        self.features = String(None, length=16)
        self.params += [self.features]


class NodeJoinAckResponse(NodeResponse):
    """
    Notification message when node (re)joined existing network again.
    Sent when a SED (re)joins the network e.g. when you reinsert the battery of a Scan

    Response to: <nothing> or NodeAddRequest
    """

    ID = b"0061"

    def __init__(self):
        super().__init__()
        # sequence number is always FFFD


class NodeAckResponse(NodeResponse):
    """
    Acknowledge message in regular format
    Sent by nodes supporting plugwise 2.4 protocol version

    Response to:
    """

    ID = b"0100"

    def __init__(self):
        super().__init__()
        self.ack_id = Int(0, 2, False)


class SenseReportResponse(NodeResponse):
    """
    Returns the current temperature and humidity of a Sense node.
    The interval this report is sent is configured by the 'SenseReportIntervalRequest' request

    Response to: <nothing>
    """

    ID = b"0105"

    def __init__(self):
        super().__init__()
        self.humidity = Int(0, length=4)
        self.temperature = Int(0, length=4)
        self.params += [self.humidity, self.temperature]


class CircleInitialRelaisStateResponse(NodeResponse):
    """
    Returns the initial relais state.

    Response to: CircleInitialRelaisStateRequest
    """

    ID = b"0139"

    def __init__(self):
        super().__init__()
        set_or_get = Int(0, length=2)
        relais = Int(0, length=2)
        self.params += [set_or_get, relais]


id_to_message = {
    b"0002": CirclePlusQueryResponse(),
    b"0003": CirclePlusQueryEndResponse(),
    b"0005": CirclePlusConnectResponse(),
    b"0006": NodeJoinAvailableResponse(),
    b"000E": NodePingResponse(),
    b"0011": StickInitResponse(),
    b"0013": CirclePowerUsageResponse(),
    b"0019": CirclePlusScanResponse(),
    b"001D": NodeRemoveResponse(),
    b"0024": NodeInfoResponse(),
    b"0027": CircleCalibrationResponse(),
    b"003A": CirclePlusRealTimeClockResponse(),
    b"003F": CircleClockResponse(),
    b"0049": CirclePowerBufferResponse(),
    b"0060": NodeFeaturesResponse(),
    b"0100": NodeAckResponse(),
    b"0105": SenseReportResponse(),
}


def get_message_response(message_id, length, seq_id):
    """
    Return message class based on sequence ID, Length of message or message ID.
    """
    # First check for known sequence ID's
    if seq_id == b"FFFD":
        return NodeJoinAckResponse()
    if seq_id == b"FFFE":
        return NodeAwakeResponse()
    if seq_id == b"FFFF":
        return NodeSwitchGroupResponse()

    # No fixed sequence ID, continue at message ID
    if message_id == b"0000":
        if length == 20:
            return NodeAckSmallResponse()
        if length == 36:
            return NodeAckLargeResponse()
        return None
    return id_to_message.get(message_id, None)



"""Plugwise Circle node object."""
from datetime import datetime, timedelta
import logging

from ..constants import (
    FEATURE_PING,
    FEATURE_POWER_CONSUMPTION_CURRENT_HOUR,
    FEATURE_POWER_CONSUMPTION_PREVIOUS_HOUR,
    FEATURE_POWER_CONSUMPTION_TODAY,
    FEATURE_POWER_CONSUMPTION_YESTERDAY,
    FEATURE_POWER_PRODUCTION_CURRENT_HOUR,
    FEATURE_POWER_USE,
    FEATURE_POWER_USE_LAST_8_SEC,
    FEATURE_RELAY,
    FEATURE_RSSI_IN,
    FEATURE_RSSI_OUT,
    MAX_TIME_DRIFT,
    MESSAGE_TIME_OUT,
    PRIORITY_HIGH,
    PRIORITY_LOW,
    PULSES_PER_KW_SECOND,
    RELAY_SWITCHED_OFF,
    RELAY_SWITCHED_ON,
)
from ..messages.requests import (
    CircleCalibrationRequest,
    CircleClockGetRequest,
    CircleClockSetRequest,
    CirclePowerBufferRequest,
    CirclePowerUsageRequest,
    CircleSwitchRelayRequest,
)
from ..messages.responses import (
    CircleCalibrationResponse,
    CircleClockResponse,
    CirclePowerBufferResponse,
    CirclePowerUsageResponse,
    NodeAckLargeResponse,
)
from ..nodes import PlugwiseNode

_LOGGER = logging.getLogger(__name__)


class PlugwiseCircle(PlugwiseNode):
    """provides interface to the Plugwise Circle nodes and base class for Circle+ nodes"""

    def __init__(self, mac, address, message_sender):
        super().__init__(mac, address, message_sender)
        self._features = (
            FEATURE_PING["id"],
            FEATURE_POWER_USE["id"],
            FEATURE_POWER_USE_LAST_8_SEC["id"],
            FEATURE_POWER_CONSUMPTION_CURRENT_HOUR["id"],
            FEATURE_POWER_CONSUMPTION_PREVIOUS_HOUR["id"],
            FEATURE_POWER_CONSUMPTION_TODAY["id"],
            FEATURE_POWER_CONSUMPTION_YESTERDAY["id"],
            FEATURE_POWER_PRODUCTION_CURRENT_HOUR["id"],
            # FEATURE_POWER_PRODUCTION_PREVIOUS_HOUR["id"],
            FEATURE_RSSI_IN["id"],
            FEATURE_RSSI_OUT["id"],
            FEATURE_RELAY["id"],
        )
        self._new_relay_state = False
        self._new_relay_stamp = datetime.now() - timedelta(seconds=MESSAGE_TIME_OUT)
        self._pulses_1s = None
        self._pulses_8s = None
        self._pulses_consumed_1h = None
        self._pulses_produced_1h = None
        self.calibration = False
        self._gain_a = None
        self._gain_b = None
        self._off_noise = None
        self._off_tot = None
        self._measures_power = True
        self.power_history = {}
        self._power_consumption_prev_hour = None
        self._power_consumption_today = None
        self._power_consumption_yesterday = None
        self.last_log_collected = False
        self.timezone_delta = datetime.now().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.utcnow().replace(minute=0, second=0, microsecond=0)
        self._clock_offset = None
        self.get_clock(self.sync_clock)
        self._request_calibration()

    @property
    def current_power_usage(self):
        """
        Returns power usage during the last second in Watts
        Based on last received power usage information
        """
        if self._pulses_1s is not None:
            return self.pulses_to_kws(self._pulses_1s) * 1000
        return None

    @property
    def current_power_usage_8_sec(self):
        """
        Returns power usage during the last 8 second in Watts
        Based on last received power usage information
        """
        if self._pulses_8s is not None:
            return self.pulses_to_kws(self._pulses_8s, 8) * 1000
        return None

    @property
    def power_consumption_current_hour(self):
        """
        Returns the power usage during this running hour in kWh
        Based on last received power usage information
        """
        if self._pulses_consumed_1h is not None:
            return self.pulses_to_kws(self._pulses_consumed_1h, 3600)
        return None

    @property
    def power_consumption_previous_hour(self):
        """Returns power consumption during the previous hour in kWh"""
        return self._power_consumption_prev_hour

    @property
    def power_consumption_today(self):
        """Total power consumption during today in kWh"""
        return self._power_consumption_today

    @property
    def power_consumption_yesterday(self):
        """Total power consumption of yesterday in kWh"""
        return self._power_consumption_yesterday

    @property
    def power_production_current_hour(self):
        """
        Returns the power production during this running hour in kWh
        Based on last received power usage information
        """
        if self._pulses_produced_1h is not None:
            return self.pulses_to_kws(self._pulses_produced_1h, 3600)
        return None

    @property
    def relay_state(self) -> bool:
        """
        Return last known relay state or the new switch state by anticipating
        the acknowledge for new state is getting in before message timeout.
        """
        if self._new_relay_stamp + timedelta(seconds=MESSAGE_TIME_OUT) > datetime.now():
            return self._new_relay_state
        return self._relay_state

    @relay_state.setter
    def relay_state(self, state):
        """Request the relay to switch state."""
        self._request_switch(state)
        self._new_relay_state = state
        self._new_relay_stamp = datetime.now()
        if state != self._relay_state:
            self.do_callback(FEATURE_RELAY["id"])

    def _request_calibration(self, callback=None):
        """Request calibration info"""
        self.message_sender(
            CircleCalibrationRequest(self._mac),
            callback,
            0,
            PRIORITY_HIGH,
        )

    def _request_switch(self, state, callback=None):
        """Request to switch relay state and request state info"""
        self.message_sender(
            CircleSwitchRelayRequest(self._mac, state),
            callback,
            0,
            PRIORITY_HIGH,
        )

    def _request_power_update(self, callback=None):
        """Request power usage and power logs of last hour"""
        if self._available:
            self.message_sender(
                CirclePowerUsageRequest(self._mac),
                callback,
            )
            # Refresh node info once an hour and request last hour power use afterwards
            if self.last_info_message < (
                datetime.now().replace(
                    minute=0,
                    second=0,
                    microsecond=0,
                )
            ):
                self._request_info(self.request_power_buffer)
            if not self.last_log_collected:
                self.request_power_buffer()

    def message_for_circle(self, message):
        """
        Process received message
        """
        if isinstance(message, CirclePowerUsageResponse):
            if self.calibration:
                self._response_power_usage(message)
                _LOGGER.debug(
                    "Power update for %s, last update %s",
                    self.mac,
                    str(self._last_update),
                )
            else:
                _LOGGER.info(
                    "Received power update for %s before calibration information is known",
                    self.mac,
                )
                self._request_calibration(self._request_power_update)
        elif isinstance(message, NodeAckLargeResponse):
            self._node_ack_response(message)
        elif isinstance(message, CircleCalibrationResponse):
            self._response_calibration(message)
        elif isinstance(message, CirclePowerBufferResponse):
            if self.calibration:
                self._response_power_buffer(message)
            else:
                _LOGGER.debug(
                    "Received power buffer log for %s before calibration information is known",
                    self.mac,
                )
                self._request_calibration(self.request_power_buffer)
        elif isinstance(message, CircleClockResponse):
            self._response_clock(message)
        else:
            self.message_for_circle_plus(message)

    def message_for_circle_plus(self, message):
        """Pass messages to PlugwiseCirclePlus class"""

    def _node_ack_response(self, message):
        """Process switch response message"""
        if message.ack_id == RELAY_SWITCHED_ON:
            if not self._relay_state:
                _LOGGER.debug(
                    "Switch relay on for %s",
                    self.mac,
                )
                self._relay_state = True
                self.do_callback(FEATURE_RELAY["id"])
        elif message.ack_id == RELAY_SWITCHED_OFF:
            if self._relay_state:
                _LOGGER.debug(
                    "Switch relay off for %s",
                    self.mac,
                )
                self._relay_state = False
                self.do_callback(FEATURE_RELAY["id"])
        else:
            _LOGGER.debug(
                "Unmanaged _node_ack_response %s received for %s",
                str(message.ack_id),
                self.mac,
            )

    def _response_power_usage(self, message):
        # Sometimes the circle returns -1 for some of the pulse counters
        # likely this means the circle measures very little power and is suffering from
        # rounding errors. Zero these out. However, negative pulse values are valid
        # for power producing appliances, like solar panels, so don't complain too loudly.

        # Power consumption last second
        if message.pulse_1s.value == -1:
            message.pulse_1s.value = 0
            _LOGGER.debug(
                "1 sec power pulse counter for node %s has value of -1, corrected to 0",
                self.mac,
            )
        self._pulses_1s = message.pulse_1s.value
        if message.pulse_1s.value != 0:
            if message.nanosecond_offset.value != 0:
                pulses_1s = (
                    message.pulse_1s.value
                    * (1000000000 + message.nanosecond_offset.value)
                ) / 1000000000
            else:
                pulses_1s = message.pulse_1s.value
            self._pulses_1s = pulses_1s
        else:
            self._pulses_1s = 0
        self.do_callback(FEATURE_POWER_USE["id"])
        # Power consumption last 8 seconds
        if message.pulse_8s.value == -1:
            message.pulse_8s.value = 0
            _LOGGER.debug(
                "8 sec power pulse counter for node %s has value of -1, corrected to 0",
                self.mac,
            )
        if message.pulse_8s.value != 0:
            if message.nanosecond_offset.value != 0:
                pulses_8s = (
                    message.pulse_8s.value
                    * (1000000000 + message.nanosecond_offset.value)
                ) / 1000000000
            else:
                pulses_8s = message.pulse_8s.value
            self._pulses_8s = pulses_8s
        else:
            self._pulses_8s = 0
        self.do_callback(FEATURE_POWER_USE_LAST_8_SEC["id"])
        # Power consumption current hour
        if message.pulse_hour_consumed.value == -1:
            message.pulse_hour_consumed.value = 0
            _LOGGER.debug(
                "1 hour consumption power pulse counter for node %s has value of -1, corrected to 0",
                self.mac,
            )
        if self._pulses_consumed_1h != message.pulse_hour_consumed.value:
            self._pulses_consumed_1h = message.pulse_hour_consumed.value
            self.do_callback(FEATURE_POWER_CONSUMPTION_CURRENT_HOUR["id"])
        # Power produced current hour
        if message.pulse_hour_produced.value == -1:
            message.pulse_hour_produced.value = 0
            _LOGGER.debug(
                "1 hour power production pulse counter for node %s has value of -1, corrected to 0",
                self.mac,
            )
        if self._pulses_produced_1h != message.pulse_hour_produced.value:
            self._pulses_produced_1h = message.pulse_hour_produced.value
            self.do_callback(FEATURE_POWER_PRODUCTION_CURRENT_HOUR["id"])

    def _response_calibration(self, message):
        """Store calibration properties"""
        for calibration in ("gain_a", "gain_b", "off_noise", "off_tot"):
            val = getattr(message, calibration).value
            setattr(self, "_" + calibration, val)
        self.calibration = True

    def pulses_to_kws(self, pulses, seconds=1):
        """
        converts the amount of pulses to kWs using the calaboration offsets
        """
        if pulses == 0 or not self.calibration:
            return 0.0
        pulses_per_s = pulses / float(seconds)
        corrected_pulses = seconds * (
            (
                (((pulses_per_s + self._off_noise) ** 2) * self._gain_b)
                + ((pulses_per_s + self._off_noise) * self._gain_a)
            )
            + self._off_tot
        )
        calc_value = corrected_pulses / PULSES_PER_KW_SECOND / seconds
        # Fix minor miscalculations
        if -0.001 < calc_value < 0.001:
            calc_value = 0.0
        return calc_value

    def request_power_buffer(self, log_address=None, callback=None):
        """Request power log of specified address"""
        if log_address is None:
            log_address = self._last_log_address
        if log_address is not None:
            if bool(self.power_history):
                # Only request last 2 power buffer logs
                self.message_sender(
                    CirclePowerBufferRequest(self._mac, log_address - 1),
                    None,
                    0,
                    PRIORITY_LOW,
                )
                self.message_sender(
                    CirclePowerBufferRequest(self._mac, log_address),
                    callback,
                    0,
                    PRIORITY_LOW,
                )
            else:
                # Collect power history info of today and yesterday
                # Each request contains 4 hours except last request
                for req_log_address in range(log_address - 13, log_address):
                    self.message_sender(
                        CirclePowerBufferRequest(self._mac, req_log_address),
                        None,
                        0,
                        PRIORITY_LOW,
                    )
                self.message_sender(
                    CirclePowerBufferRequest(self._mac, log_address),
                    callback,
                    0,
                    PRIORITY_LOW,
                )

    def _response_power_buffer(self, message):
        """returns information about historical power usage
        each response contains 4 log buffers and each log buffer contains data for 1 hour
        """
        if message.logaddr.value == self._last_log_address:
            self.last_log_collected = True
        # Collect logged power usage
        for i in range(1, 5):
            if getattr(message, "logdate%d" % (i,)).value is not None:
                log_date = getattr(message, "logdate%d" % (i,)).value
                if getattr(message, "pulses%d" % (i,)).value == 0:
                    self.power_history[log_date] = 0.0
                else:
                    self.power_history[log_date] = self.pulses_to_kws(
                        getattr(message, "pulses%d" % (i,)).value, 3600
                    )
        # Cleanup history for more than 2 day's ago
        if len(self.power_history.keys()) > 48:
            for log_date in list(self.power_history.keys()):
                if (log_date + self.timezone_delta - timedelta(hours=1)).date() < (
                    datetime.now().today().date() - timedelta(days=1)
                ):
                    del self.power_history[log_date]
        # Recalculate power use counters
        last_hour_usage = 0
        today_power = 0
        yesterday_power = 0
        for log_date in self.power_history:
            if (log_date + self.timezone_delta) == datetime.now().today().replace(
                minute=0, second=0, microsecond=0
            ):
                last_hour_usage = self.power_history[log_date]
            if (
                log_date + self.timezone_delta - timedelta(hours=1)
            ).date() == datetime.now().today().date():
                today_power += self.power_history[log_date]
            if (log_date + self.timezone_delta - timedelta(hours=1)).date() == (
                datetime.now().today().date() - timedelta(days=1)
            ):
                yesterday_power += self.power_history[log_date]
        if self._power_consumption_prev_hour != last_hour_usage:
            self._power_consumption_prev_hour = last_hour_usage
            self.do_callback(FEATURE_POWER_CONSUMPTION_PREVIOUS_HOUR["id"])
        if self._power_consumption_today != today_power:
            self._power_consumption_today = today_power
            self.do_callback(FEATURE_POWER_CONSUMPTION_TODAY["id"])
        if self._power_consumption_yesterday != yesterday_power:
            self._power_consumption_yesterday = yesterday_power
            self.do_callback(FEATURE_POWER_CONSUMPTION_YESTERDAY["id"])

    def _response_clock(self, message):
        log_date = datetime(
            datetime.now().year,
            datetime.now().month,
            datetime.now().day,
            message.time.value.hour,
            message.time.value.minute,
            message.time.value.second,
        )
        clock_offset = message.timestamp.replace(microsecond=0) - (
            log_date + self.timezone_delta
        )
        if clock_offset.days == -1:
            self._clock_offset = clock_offset.seconds - 86400
        else:
            self._clock_offset = clock_offset.seconds
        _LOGGER.debug(
            "Clock of node %s has drifted %s sec",
            self.mac,
            str(self._clock_offset),
        )

    def get_clock(self, callback=None):
        """get current datetime of internal clock of Circle."""
        self.message_sender(
            CircleClockGetRequest(self._mac),
            callback,
            0,
            PRIORITY_LOW,
        )

    def set_clock(self, callback=None):
        """set internal clock of CirclePlus."""
        self.message_sender(
            CircleClockSetRequest(self._mac, datetime.utcnow()),
            callback,
        )

    def sync_clock(self, max_drift=0):
        """Resync clock of node if time has drifted more than MAX_TIME_DRIFT"""
        if self._clock_offset is not None:
            if max_drift == 0:
                max_drift = MAX_TIME_DRIFT
            if (self._clock_offset > max_drift) or (self._clock_offset < -(max_drift)):
                _LOGGER.info(
                    "Reset clock of node %s because time has drifted %s sec",
                    self.mac,
                    str(self._clock_offset),
                )
                self.set_clock()

    ## TODO: All functions below can be removed when HA component is changed to use the property values ##
    def get_relay_state(self) -> bool:
        """Return last known relay state."""
        _LOGGER.warning(
            "Function 'get_relay_state' will be removed in future, use the 'relay_state' property instead !",
        )
        return self._relay_state

    def set_relay_state(self, state: bool, callback=None):
        """Switch relay."""
        _LOGGER.warning(
            "Function 'set_relay_state' will be removed in future, use the 'relay_state' property instead !",
        )
        self._request_switch(state, callback)

    def get_power_usage(self):
        """
        Returns power usage during the last second in Watts
        Based on last received power usage information
        """
        _LOGGER.warning(
            "Function 'get_power_usage' will be removed in future, use the 'current_power_usage' property instead !",
        )
        if self._pulses_1s is not None:
            return self.pulses_to_kws(self._pulses_1s) * 1000
        return None

    def get_power_usage_8_sec(self):
        """
        Returns power usage during the last 8 second in Watts
        Based on last received power usage information
        """
        _LOGGER.warning(
            "Function 'get_power_usage_8_sec' will be removed in future, use the 'current_power_usage_8_sec' property instead !",
        )
        if self._pulses_8s is not None:
            return self.pulses_to_kws(self._pulses_8s, 8) * 1000
        return None

    def get_power_consumption_current_hour(self):
        """
        Returns the power usage during this running hour in kWh
        Based on last received power usage information
        """
        _LOGGER.warning(
            "Function 'get_power_consumption_current_hour' will be removed in future, use the 'power_consumption_current_hour' property instead !",
        )
        if self._pulses_consumed_1h is not None:
            return self.pulses_to_kws(self._pulses_consumed_1h, 3600)
        return None

    def get_power_production_current_hour(self):
        """
        Returns the power production during this running hour in kWh
        Based on last received power usage information
        """
        _LOGGER.warning(
            "Function 'get_power_production_current_hour' will be removed in future, use the 'power_production_current_hour' property instead !",
        )
        if self._pulses_produced_1h is not None:
            return self.pulses_to_kws(self._pulses_produced_1h, 3600)
        return None

    def get_power_consumption_prev_hour(self):
        """Returns power consumption during the previous hour in kWh"""
        _LOGGER.warning(
            "Function 'get_power_consumption_prev_hour' will be removed in future, use the 'power_consumption_previous_hour' property instead !",
        )
        return self._power_consumption_prev_hour

    def get_power_consumption_today(self):
        """Total power consumption during today in kWh"""
        _LOGGER.warning(
            "Function 'get_power_consumption_today' will be removed in future, use the 'power_consumption_today' property instead !",
        )
        return self._power_consumption_today

    def get_power_consumption_yesterday(self):
        """Total power consumption of yesterday in kWh"""
        _LOGGER.warning(
            "Function 'get_power_consumption_yesterday' will be removed in future, use the 'power_consumption_yesterday' property instead !",
        )
        return self._power_consumption_yesterday



import codecs
import os.path

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name="plugwise",
    version=get_version("plugwise/__init__.py"),
    description="Plugwise (Adam/Anna/P1/Stick/Stretch) API to use in conjunction with Home Assistant Core.",
    long_description="Plugwise API to use in conjunction with Home Assistant, but it can also be used as a python-module.",
    keywords="HomeAssistant HA Home Assistant Anna Adam P1 Smile Stretch Stick Plugwise",
    url="https://github.com/plugwise/python-plugwise",
    author="Plugwise-team",
    author_email="info@compa.nl",
    license="MIT",
    packages=find_packages(),
    install_requires=[
        "aiohttp",
        "async_timeout<4.0",
        "crcmod",
        "defusedxml",
        "munch",
        "pyserial",
        "pytz",
        "python-dateutil",
        "semver",
    ],
    zip_safe=False,
)
